"""pypyr.steps.cmd unit tests.

For tests involving stdout & stderr set to files, see:
tests/integration/pypyr/steps/cmd_int_test.py
"""
from pathlib import Path
import subprocess

import pytest

from pypyr.config import config
from pypyr.context import Context
from pypyr.errors import ContextError, KeyNotInContextError
import pypyr.steps.cmd

cmd_path = Path.cwd().joinpath('tests/testfiles/cmds')
is_windows = config.is_windows


def get_cmd(posix, win):
    """Return posix or win depending on current platform."""
    return win if is_windows else posix


def test_cmd_str_with_args():
    """Single command string works with string interpolation."""
    cmd = get_cmd('tests/testfiles/cmds/args.sh',
                  r'tests\testfiles\cmds\args.bat')
    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd,
                       'cmd': '{d} {a} "{b}" {c}'})
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_str_with_spaces_in_path():
    """Single command string works with spaces in path."""
    cmd = get_cmd('tests/testfiles/cmds/"file with space.sh"',
                  r'tests\testfiles\cmds\file with space.bat')
    context = Context({'a': 'one two',
                       'd': cmd,
                       'cmd': '{d} "{a}"'})
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context

    # and with quotes
    if is_windows:
        context = Context({
            'cmd': r'"tests\testfiles\cmds\file with space.bat" "one two"'})
        pypyr.steps.cmd.run_step(context)

        assert 'cmdOut' not in context


def test_cmd_str_with_args_rel_path_with_backslash():
    """On Windows can use relative path if backslash in path."""
    if is_windows:
        cmd = r'tests\testfiles\cmds\args.bat'
    else:
        cmd = 'tests/testfiles/cmds/args.sh'

    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd,
                       'cmd': '{d} {a} "{b}" {c}'})
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_dict_input_with_args():
    """Single command string works with multiple args."""
    cmd = get_cmd('tests/testfiles/cmds/args.sh',
                  r'tests\testfiles\cmds\args.bat')
    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd,
                       'cmd': {
                           'run': '{d} {a} "{b}" {c}'}})
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_dict_input_with_string_interpolation_save_out():
    """Single command string works with string interpolation."""
    cmd = get_cmd('echo blah',
                  r'tests\testfiles\cmds\echo.bat blah')
    context = Context({'fileName': 'deleteinterpolatedme.arb',
                       'cmd': {'run': cmd,
                               'save': True}})
    pypyr.steps.cmd.run_step(context)

    # dict like accessors only here for backwards compatibility
    assert context['cmdOut']['returncode'] == 0
    assert context['cmdOut']['stdout'] == 'blah'
    assert not context['cmdOut']['stderr']

    assert context['cmdOut'].returncode == 0
    assert context['cmdOut'].stdout == 'blah'
    assert not context['cmdOut'].stderr


def test_cmd_dict_input_sequence_with_cwd():
    """Single command string works with cwd."""
    cmd = get_cmd('pwd', str(cmd_path.joinpath('pwd.bat')))

    context = Context({'fileName': 'deleteinterpolatedme.arb',
                       'cmd': {'run': cmd,
                               'save': True,
                               'cwd': 'tests'}})
    pypyr.steps.cmd.run_step(context)

    assert context['cmdOut'].returncode == 0
    assert Path(context['cmdOut'].stdout).samefile('./tests')
    assert not context['cmdOut'].stderr


def test_cmd_dict_input_sequence_with_cwd_interpolate():
    """Single command string works with cwd interpolation."""
    if is_windows:
        # windows path supports front slashes IF it's absolute
        cmd = cmd_path.joinpath('pwd.bat').as_posix()
    else:
        # deliberately testing relative path resolution here
        cmd = 'testfiles/cmds/pwd.sh'

    context = Context({'k1': 'tests',
                       'cmd': {'run': cmd,
                               'save': True,
                               'cwd': '{k1}'}})
    pypyr.steps.cmd.run_step(context)

    assert context['cmdOut'].returncode == 0
    assert Path(context['cmdOut'].stdout).samefile('./tests')
    assert not context['cmdOut'].stderr


def test_cmd_error_throws():
    """Process returning 1 should throw CalledProcessError with no save."""
    cmd = get_cmd('tests/testfiles/cmds/exit1.sh',
                  r'tests\testfiles\cmds\exit1.bat')
    with pytest.raises(subprocess.CalledProcessError):
        context = Context({'cmd': cmd})
        pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_error_throws_with_save_true():
    """Process returning 1 should throw CalledProcessError."""
    cmd = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                  r'tests\testfiles\cmds\exitwitherr.bat')

    with pytest.raises(subprocess.CalledProcessError):
        context = Context({'cmd': {'run': cmd, 'save': True}})
        pypyr.steps.cmd.run_step(context)

    assert context['cmdOut'].returncode == 1
    assert not context['cmdOut'].stdout
    assert context['cmdOut'].stderr == 'arb err here'


def test_cmd_context_cmd_throw():
    """Empty cmd in context should throw assert error."""
    with pytest.raises(KeyNotInContextError) as err_info:
        context = Context({'blah': 'blah blah'})
        pypyr.steps.cmd.run_step(context)

    assert str(err_info.value) == ("context['cmd'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.cmd.")


def test_cmd_run_save_with_stdout_stderr():
    """Raise err when setting stdout | stderr alongside save."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two three',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2],
            'save': True,
            'stdout': '/arb1',
            'stderr': '/arb2'}
    })

    with pytest.raises(ContextError) as err:
        pypyr.steps.cmd.run_step(context)

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")


def test_cmd_run_save_with_stdout_or_stderr():
    """Raise err when setting stdout | stderr alongside save - or not and."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two three',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2],
            'save': True,
            'stderr': '/arb2'}
    })

    with pytest.raises(ContextError) as err:
        pypyr.steps.cmd.run_step(context)

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")


def test_cmd_error_throws_with_save_true_executable_not_found():
    """Raise error when can't find the subprocess executable."""
    cmd = get_cmd('tests/testfiles/cmds/xxx',
                  r'tests\testfiles\cmds\xxx')

    with pytest.raises(FileNotFoundError) as err:
        context = Context({'cmd': {'run': cmd, 'save': True}})
        pypyr.steps.cmd.run_step(context)

    # cmdOut NOT populated because couldn't run subproc at all.
    assert 'cmdOut' not in context

    # on windows, FileNotFoundError does NOT set filename, instead:
    # FileNotFoundError(2, 'The system cannot find the file specified',
    #                   None, 2 None)
    if not is_windows:
        err.value.filename == cmd

# region list input


def test_cmd_run_has_list_input_with_complex_args():
    """List input to run complex dict works with multiple args."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': {
                           'run': [
                               '{d} {a} "{b}" {c}',
                               '{e} four "five six" seven'],
                            'save': False}
                       })
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_list_input_with_simple_cmd_strings():
    """List input with command string works with multiple args."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': [
                           '{d} {a} "{b}" {c}',
                           '{e} four "five six" seven']})
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_list_input_with_complex_args():
    """List input mixing command string & dict works with multiple args."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': [
                           {'run': '{d} {a} "{b}" {c}',
                            'save': False},
                           '{e} four "five six" seven']})
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


def test_cmd_list_input_with_simple_cmd_strings_error_on_first():
    """List input with command string works stops on 1st error."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('echo blah',
                   r'tests\testfiles\cmds\echo.bat blah')

    context = Context({'a': 'WRONG VALUE',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': [
                           '{d} {a} "{b}" {c}',
                           '{e}']})

    with pytest.raises(subprocess.CalledProcessError) as err:
        pypyr.steps.cmd.run_step(context)

    if is_windows:
        assert err.value.cmd == f'{cmd1} WRONG VALUE "two two" three'
    else:
        assert err.value.cmd == [cmd1, 'WRONG', 'VALUE', 'two two', 'three']
    assert 'cmdOut' not in context


def test_cmd_run_has_list_input_with_complex_args_error_on_first():
    """List input to run complex dict works with error on first."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'WRONG',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': {
                           'run': [
                               '{d} {a} "{b}" {c}',
                               '{e} four "five six" seven'],
                            'save': False}
                       })

    with pytest.raises(subprocess.CalledProcessError) as err:
        pypyr.steps.cmd.run_step(context)

    if is_windows:
        assert err.value.cmd == f'{cmd1} WRONG "two two" three'
    else:
        assert err.value.cmd == [cmd1, 'WRONG', 'two two', 'three']
    assert 'cmdOut' not in context


def test_cmd_list_input_with_complex_args_error_on_first():
    """List input mixing command string & dict stops on 1st."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'WRONG',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': [
                           {'run': '{d} {a} "{b}" {c}',
                            'save': False},
                           '{e} four "five six" seven']})

    with pytest.raises(subprocess.CalledProcessError) as err:
        pypyr.steps.cmd.run_step(context)

    if is_windows:
        assert err.value.cmd == f'{cmd1} WRONG "two two" three'
    else:
        assert err.value.cmd == [cmd1, 'WRONG', 'two two', 'three']
    assert 'cmdOut' not in context


def test_cmd_list_input_with_complex_args_error_on_first_save():
    """List input mixing command string & dict stops on 1st save output."""
    cmd1 = get_cmd('tests/testfiles/cmds/args.sh',
                   r'tests\testfiles\cmds\args.bat')

    cmd2 = get_cmd('tests/testfiles/cmds/args2.sh',
                   r'tests\testfiles\cmds\args2.bat')

    context = Context({'a': 'WRONG',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd1,
                       'e': cmd2,
                       'cmd': [
                           {'run': '{d} {a} "{b}" {c}',
                            'save': True},
                           '{e} WRONG "five six" seven']})

    with pytest.raises(subprocess.CalledProcessError) as err:
        pypyr.steps.cmd.run_step(context)

    if is_windows:
        assert err.value.cmd == f'{cmd1} WRONG "two two" three'
    else:
        assert err.value.cmd == [cmd1, 'WRONG', 'two two', 'three']
    out = context['cmdOut']
    assert out.returncode == 1
    assert out.stdout == ''
    assert out.stderr == 'assert failed'


def test_cmd_run_has_list_input_save():
    """List input to run complex dict save all output."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo "two three"',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2],
            'save': True}
    })
    pypyr.steps.cmd.run_step(context)

    out = context['cmdOut']
    assert len(out) == 2
    out1 = out[0]
    assert out1.returncode == 0
    assert out1.stdout == 'one'
    assert not out1.stderr

    out2 = out[1]
    assert out2.returncode == 0
    assert out2.stdout == 'two three'
    assert not out2.stderr


def test_cmd_run_has_list_and_run_with_list_save():
    """List input with run also a list, everything save to results."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two',
                   r'tests\testfiles\cmds\echo.bat two')

    cmd3 = get_cmd('echo three',
                   r'tests\testfiles\cmds\echo.bat three')

    cmd4 = get_cmd('echo four',
                   r'tests\testfiles\cmds\echo.bat four')

    cmd5 = get_cmd('echo five',
                   r'tests\testfiles\cmds\echo.bat five')

    context = Context({
        'cmd': [
            {'run': [cmd1, cmd2],
             'save': True},
            {'run': [cmd3, cmd4],
             'save': True},
            cmd5,
        ]
    })
    pypyr.steps.cmd.run_step(context)

    out = context['cmdOut']
    assert len(out) == 4

    out1 = out[0]
    assert out1.returncode == 0
    assert out1.stdout == 'one'
    assert not out1.stderr

    out2 = out[1]
    assert out2.returncode == 0
    assert out2.stdout == 'two'
    assert not out2.stderr

    out3 = out[2]
    assert out3.returncode == 0
    assert out3.stdout == 'three'
    assert not out3.stderr

    out4 = out[3]
    assert out4.returncode == 0
    assert out4.stdout == 'four'
    assert not out4.stderr


def test_cmd_run_has_list_input_save_dev_null():
    """List input to run complex dict save all output to /dev/null."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two three',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2],
            'stdout': '/dev/null',
            'stderr': '/dev/null'}
    })
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context


# endregion list input

# region encoding

powershell = "powershell -ExecutionPolicy Bypass -NoLogo -NoProfile"


def test_cmd_run_save_with_encoding():
    """Encoding works."""
    if is_windows:
        cmd = f"{powershell} ./tests/testfiles/cmds/bytes-out.ps1"
        encoding = 'utf-16le'
    else:
        cmd = 'tests/testfiles/cmds/bytes-out.sh'
        encoding = 'utf-16'

    context = Context({
        'cmd': {
            'run': cmd,
            'save': True,
            'encoding': encoding}
    })
    pypyr.steps.cmd.run_step(context)

    out = context['cmdOut']
    assert out.returncode == 0
    assert out.stdout == 'one two three'
    assert out.stderr == 'four'


def test_cmd_run_save_with_binary_mode():
    """Binary mode does not strip lf."""
    if is_windows:
        # the literal \ escaping the " is NB
        ps_cmd = """\
$out=[System.Text.Encoding]::Unicode.GetBytes(\\"one`n\\");
[System.Console]::OpenStandardOutput().Write($out, 0, $out.Length);
$err=[System.Text.Encoding]::Unicode.GetBytes(\\"two`n\\");
[System.Console]::OpenStandardError().Write($err, 0, $err.Length);"""

        cmd = f'powershell -NoLogo -NoProfile -Command {ps_cmd}'
    else:
        cmd = 'tests/testfiles/cmds/bytes-out-lf.sh'

    context = Context({
        'cmd': {
            'run': cmd,
            'save': True,
            'bytes': True}
    })
    pypyr.steps.cmd.run_step(context)

    out = context['cmdOut']
    assert out.returncode == 0
    assert out.stdout.decode('utf-16') == 'one\n'
    assert out.stderr.decode('utf-16') == 'two\n'

# endregion encoding



"""Nested calls inside loops. These pipelines are in ./tests/pipelines/."""
from pypyr import pipelinerunner


def test_switch_nested():
    """Nested switch control-of-flow works."""
    out = pipelinerunner.run('tests/pipelines/switch/nested')

    assert out == {'out': [
        'begin',
        'A.1',
        'B',
        'A.2',
        'C',
        'A.3',
        'end.'
    ]}



"""Test custom loader which returns directly the same thing that was passed."""


def get_pipeline_definition(pipeline_name, parent):
    """Return inputs as a mock pipeline loader stub."""
    return {'pipeline_name': pipeline_name, 'parent': parent}



"""Create a standard python venv using the stdlib venv module."""
import logging

from pypyr.steps.dsl.venv import VenvCreatorStep

logger = logging.getLogger(__name__)


def run_step(context):
    """Create a stdlib venv.

    Args:
        context (pypyr.context.Context): Context must contain
            venv: path

            or

            venv:
                - path1
                - path2

            or

            venv:
                path: path1
                pip: dep1 dep2 -e dep3

            or

            venv:
                path:
                    - path1
                    - path2
                pip: dep1 dep2 -e dep3

            or you can mix simple string and extended mapping syntax:

            venv:
                - path1

                - path: path2
                  pip: dep1 dep2

                - path:
                    - ./path3
                    - /path4
                  pip: dep3 dep4
    """
    logger.debug("started")
    VenvCreatorStep.from_context(context).run_step()
    logger.debug("done")



"""append.py unit tests."""
import pytest

from pypyr.context import Context
from pypyr.dsl import PyString
from pypyr.errors import (ContextError,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)
from pypyr.steps import append


# region validation
def test_append_no_input():
    """Context append must exist."""
    with pytest.raises(KeyNotInContextError):
        append.run_step(Context())


def test_append_not_a_dict():
    """Context append must be a dict."""
    with pytest.raises(ContextError) as err:
        append.run_step(Context({
            'append': 1}))

    assert str(err.value) == (
        "context['append'] must exist, be iterable and contain 'list' for "
        "pypyr.steps.append. argument of type 'int' is not iterable")


def test_append_no_base_object():
    """List input must exist."""
    with pytest.raises(KeyNotInContextError) as err:
        append.run_step(Context({
            'append': {}}))

    assert str(err.value) == (
        "context['append']['list'] doesn't exist. It must exist for "
        "pypyr.steps.append.")


def test_append_no_add_me():
    """Input add_me must exist."""
    with pytest.raises(KeyNotInContextError):
        append.run_step(Context({
            'append': {
                'list': 'arblist'
            }}))


def test_append_with_empty_base():
    """Input list is empty."""
    with pytest.raises(KeyInContextHasNoValueError):
        append.run_step(Context({
            'append': {
                        'list': '',
                        'addMe': ''}}))


def test_append_with_none_base():
    """Input list is empty."""
    with pytest.raises(KeyInContextHasNoValueError):
        append.run_step(Context({
            'append': {
                        'list': None,
                        'addMe': None}}))


# endregion validation

def test_append_with_none_add():
    """Add None to list."""
    context = Context({
        'append': {
            'list': 'arblist',
            'addMe': None
        }})

    append.run_step(context)

    context['append']['addMe'] = 'one'

    append.run_step(context)

    assert context['arblist'] == [None, 'one']
    assert len(context) == 2


def test_append_pass_with_minimal_create():
    """Create list with minimal parameters success."""
    context = Context({
        'append': {
            'list': 'arblist',
            'addMe': 1
        }})

    append.run_step(context)

    context['append']['addMe'] = 2

    append.run_step(context)

    assert context['arblist'] == [1, 2]
    assert len(context) == 2


def test_append_pass_with_create_from_list_no_extend():
    """Minimal paramaters success."""
    context = Context({
        'append': {
            'list': 'arblist',
            'addMe': [1, 2]
        }})

    append.run_step(context)

    context['append']['addMe'] = 3

    append.run_step(context)

    assert context['arblist'] == [[1, 2], 3]
    assert len(context) == 2


def test_append_pass_with_create_from_list_extend():
    """Create new list with extend."""
    context = Context({
        'append': {
            'list': 'arblist',
            'addMe': [1, 2],
            'unpack': True
        }})

    append.run_step(context)

    context['append']['unpack'] = False
    context['append']['addMe'] = 3

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3]
    assert len(context) == 2


def test_append_pass_existing():
    """Append existing list."""
    context = Context({
        'arblist': [1, 2],
        'append': {
            'list': 'arblist',
            'addMe': 3
        }})

    append.run_step(context)

    context['append']['addMe'] = 4

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3, 4]
    assert len(context) == 2


def test_append_pass_with_extend_existing():
    """Extend existing list."""
    context = Context({
        'arblist': [1, 2],
        'append': {
            'list': 'arblist',
            'addMe': [3, 4],
            'unpack': True
        }})

    append.run_step(context)

    context['append']['unpack'] = False
    context['append']['addMe'] = 5

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3, 4, 5]
    assert len(context) == 2


def test_append_with_list_input():
    """Input is a list, not a string key."""
    context = Context({
        'arblist': [1, 2],
        'append': {
            'list': PyString('arblist'),
            'addMe': 3
        }})

    append.run_step(context)

    context['append']['addMe'] = 4

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3, 4]
    assert len(context) == 2


def test_append_with_list_input_extend():
    """Input is a list, not a string key with an extend."""
    context = Context({
        'arblist': [1, 2],
        'append': {
            'list': PyString('arblist'),
            'addMe': 3
        }})

    append.run_step(context)

    context['append']['addMe'] = [4, 5]
    context['append']['unpack'] = True

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3, 4, 5]
    assert len(context) == 2


def test_append_with_formatting():
    """Use formatting expressions."""
    context = Context({
        'arblist': [1, 2],
        'addthis': 3,
        'append': {
            'list': PyString('arblist'),
            'addMe': '{addthis}'
        }})

    append.run_step(context)

    context['append']['addMe'] = 4

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3, 4]
    assert len(context) == 3


def test_append_list_with_formatting_no_extend():
    """Use formatting expressions to append list without extend."""
    context = Context({
        'arblist': [1, 2],
        'addthis': [3, 4],
        'append': {
            'list': PyString('arblist'),
            'addMe': '{addthis}'
        }})

    append.run_step(context)

    context['append']['addMe'] = 5

    append.run_step(context)

    assert context['arblist'] == [1, 2, [3, 4], 5]
    assert len(context) == 3


def test_append_with_formatting_extend():
    """Use formatting expressions."""
    context = Context({
        'arblist': [1, 2],
        'addthis': [3, 4],
        'extend_me': True,
        'append': {
            'list': PyString('arblist'),
            'addMe': '{addthis}',
            'unpack': '{extend_me}'
        }})

    append.run_step(context)

    context['append']['addMe'] = [5, 6]

    append.run_step(context)

    assert context['arblist'] == [1, 2, 3, 4, 5, 6]
    assert len(context) == 4


def test_append_with_strings():
    """Append strings."""
    context = Context({
        'arblist': [1, 2],
        'append': {
            'list': PyString('arblist'),
            'addMe': 'three'
        }})

    append.run_step(context)

    context['append']['addMe'] = 'four'

    append.run_step(context)

    assert context['arblist'] == [1, 2, 'three', 'four']
    assert len(context) == 2


def test_append_with_strings_unpack():
    """Append strings with unpack."""
    context = Context({
        'arblist': [1, 2],
        'append': {
            'list': PyString('arblist'),
            'addMe': 'xy',
            'unpack': True
        }})

    append.run_step(context)

    context['append']['addMe'] = 'z'

    append.run_step(context)

    assert context['arblist'] == [1, 2, 'x', 'y', 'z']
    assert len(context) == 2



"""list.py unit tests."""
import logging
import pypyr.parser.list
from tests.common.utils import patch_logger


def test_list_args_parses_to_dict():
    """Comma delimited input returns dictionary key argList with value list."""
    out = pypyr.parser.list.get_parsed_context(['value 1',
                                                'value 2',
                                                ' value3'])
    assert out['argList'] == ['value 1', 'value 2', ' value3']
    assert len(out) == 1, "1 item expected"
    assert len(out['argList']) == 3, "3 items expected in argList"


def test_list__arg_parses_to_single_entry():
    """Single entry should return list with 1 item."""
    out = pypyr.parser.list.get_parsed_context(['value 1 value 2 value3'])
    assert out['argList'] == ['value 1 value 2 value3']
    assert len(out) == 1, "1 item expected in context"
    assert len(out['argList']) == 1, "1 item expected in argList"


def test_empty_args_empty_list():
    """Empty input args should return empty list."""
    with patch_logger(
            'pypyr.parser.list', logging.DEBUG) as mock_logger_debug:
        out = pypyr.parser.list.get_parsed_context(None)

    assert not out['argList']
    assert out['argList'] == []
    mock_logger_debug.assert_called_with(
        "pipeline invoked without context arg set. For "
        "this list parser you're looking for something like:\n"
        "pypyr pipelinename spam eggs\n"
        "OR: pypyr pipelinename spam.")


def test_builtin_list_still_works():
    """Don't break built-in list keyword."""
    test_list = [0, 1, 2]
    test_list.append(3)
    assert test_list.count(0) == 1
    test_list.extend([4, 5, 6])
    assert test_list == [0, 1, 2, 3, 4, 5, 6]
    assert isinstance(test_list, list)



"""pype.py unit tests."""
import logging
from unittest.mock import call, Mock

import pytest

from pypyr.context import Context
from pypyr.errors import (
    ContextError,
    Stop,
    KeyInContextHasNoValueError,
    KeyNotInContextError)

from pypyr.pipeline import Pipeline
from pypyr.pipedef import PipelineDefinition, PipelineInfo
import pypyr.steps.pype as pype

from tests.common.utils import patch_logger

# region fixtures


def new_pipe_and_args_wrapper(mock_pipe):
    """Return ref to Pipeline new_pipe_and_args, overriding the cls arg.

    Reason is normal patch does not patch out the cls arg on a class method.

    Intercept the factory method, passing in a mock as type to instantiate.

    Args:
        mock_pipe: Replace the 1st arg (i.e cls) to the classmethod with me.
    """
    # get the reference to the underlying method before it's patched
    og_ref = Pipeline.new_pipe_and_args.__func__

    def new_pipe_and_args(*args, **kwargs):
        # the first arg is cls - for which we're substituting the mock
        return og_ref(mock_pipe, *args, **kwargs)

    return new_pipe_and_args


@pytest.fixture
def mock_pipe(monkeypatch):
    """Intercept Pipeline.new_pipe_and_args factory method."""
    mock_pipe = Mock(spec=Pipeline)
    mock_pipe._get_parse_input = Pipeline._get_parse_input

    monkeypatch.setattr('pypyr.steps.pype.Pipeline.new_pipe_and_args',
                        new_pipe_and_args_wrapper(mock_pipe))

    return mock_pipe


# endregion fixtures


def get_arb_pipeline_scope(context):
    """Context must be in pipeline scope to get current pipe info."""
    pipeline = Pipeline('arb pipe')
    pipeline.pipeline_definition = PipelineDefinition(
        pipeline=None,
        info=PipelineInfo(pipeline_name='arb',
                          loader=None,
                          parent=None))
    return context.pipeline_scope(pipeline)

# region get_arguments


def test_pype_get_arguments_all():
    """Parse all input from context."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'},
            'out': 'out value',
            'pipeArg': 'argument here',
            'useParentContext': False,
            'skipParse': 'skip parse',
            'raiseError': 'raise err',
            'loader': 'test loader',
            'groups': ['gr'],
            'success': 'sg',
            'failure': 'fg',
            'pyDir': 'arb/dir',
            'parent': 'the parent'
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args == {'a': 'b'}
    assert out == 'out value'
    assert not use_parent_context
    assert skip_parse == 'skip parse'
    assert pipe_arg == ['argument', 'here']
    assert raise_error == 'raise err'
    assert loader == 'test loader'
    assert groups == ['gr']
    assert success_group == 'sg'
    assert failure_group == 'fg'
    assert py_dir == 'arb/dir'
    assert parent == 'the parent'


def test_pype_get_arguments_all_with_interpolation():
    """Parse all input from context."""
    context = Context({
        'pipeName': 'pipe name',
        'argsHere': {'a': '{pipeName}'},
        'outHere': 'out here',
        'argHere': 'argument here',
        'parentContext': False,
        'skipParse': 'skip parse',
        'raiseErr': 'raise err',
        'loaderHere': 'test loader',
        'groups': ['gr'],
        'success': 'sg',
        'failure': 'fg',
        'dir': 'arb/dir',
        'parent': 'the parent',
        'pype': {
            'name': '{pipeName}',
            'args': '{argsHere}',
            'out': '{outHere}',
            'pipeArg': '{argHere}',
            'useParentContext': '{parentContext}',
            'skipParse': '{skipParse}',
            'raiseError': '{raiseErr}',
            'loader': '{loaderHere}',
            'groups': '{groups}',
            'success': '{success}',
            'failure': '{failure}',
            'pyDir': '{dir}',
            'parent': '{parent}'
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args == {'a': 'pipe name'}
    assert out == 'out here'
    assert not use_parent_context
    assert pipe_arg == ['argument', 'here']
    assert skip_parse == 'skip parse'
    assert raise_error == 'raise err'
    assert loader == 'test loader'
    assert groups == ['gr']
    assert success_group == 'sg'
    assert failure_group == 'fg'
    assert py_dir == 'arb/dir'
    assert parent == 'the parent'


def test_pype_get_arguments_defaults():
    """Parse all input from context and assign defaults where not specified."""
    context = Context({
        'pype': {
            'name': 'pipe name'
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args is None
    assert out is None
    assert use_parent_context
    assert isinstance(use_parent_context, bool)
    assert pipe_arg is None
    assert skip_parse
    assert isinstance(skip_parse, bool)
    assert raise_error
    assert isinstance(raise_error, bool)
    assert loader is None
    assert groups is None
    assert success_group is None
    assert failure_group is None
    assert py_dir is None
    # defaults to current pipeline's parent if cascading loader.
    assert parent is None


def test_pype_get_arguments_missing_pype():
    """Missing pype throw."""
    context = Context()

    with pytest.raises(KeyNotInContextError) as err_info:
        pype.get_arguments(context)

    assert str(err_info.value) == ("context['pype'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.pype.")


def test_pype_get_args_not_a_dict():
    """When args not a dict raise."""
    context = Context({'pype': {'name': 'blah', 'args': 'arb'}})

    with pytest.raises(ContextError) as err_info:
        pype.get_arguments(context)

    assert str(err_info.value) == (
        "pypyr.steps.pype 'args' in the 'pype' context item "
        "must be a dict.")


def test_pype_get_out_set_with_use_parent_context():
    """When out is present useParentContext must be false."""
    context = Context({'pype': {'name': 'blah',
                                'out': 'arb',
                                'useParentContext': True}})

    with pytest.raises(ContextError) as err_info:
        pype.get_arguments(context)

    assert str(err_info.value) == (
        "pypyr.steps.pype pype.out is only "
        "relevant if useParentContext = False. If you're using the parent "
        "context, no need to have out args since their values will already be "
        "in context. If you're NOT using parent context and you've specified "
        "pype.args, just leave off the useParentContext key and it'll default "
        "to False under the hood, or set it to False yourself if you keep it "
        "in.")


def test_pype_get_arguments_missing_name():
    """Missing pype name throw."""
    context = Context({'pype': {}})

    with pytest.raises(KeyNotInContextError) as err_info:
        pype.get_arguments(context)

    assert str(err_info.value) == (
        "pypyr.steps.pype missing 'name' in the 'pype' "
        "context item. You need to specify the pipeline name to run another "
        "pipeline.")


def test_pype_get_arguments_name_empty():
    """Empty pype name throw."""
    context = Context({'pype': {'name': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        pype.get_arguments(context)

    assert str(err_info.value) == ("pypyr.steps.pype ['pype']['name'] exists "
                                   "but is empty.")


def test_pype_get_arguments_group_str():
    """Parse group as str input from context."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'groups': 'gr',
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args is None
    assert out is None
    assert use_parent_context
    assert isinstance(use_parent_context, bool)
    assert pipe_arg is None
    assert skip_parse
    assert isinstance(skip_parse, bool)
    assert raise_error
    assert isinstance(raise_error, bool)
    assert loader is None
    assert groups == ['gr']
    assert success_group is None
    assert failure_group is None
    assert py_dir is None
    assert parent is None


def test_pype_get_arguments_group_str_interpolate():
    """Parse group as interpolated str input from context."""
    context = Context({
        'group': 'gr',
        'pype': {
            'name': 'pipe name',
            'groups': '{group}',
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args is None
    assert out is None
    assert use_parent_context
    assert isinstance(use_parent_context, bool)
    assert pipe_arg is None
    assert skip_parse
    assert isinstance(skip_parse, bool)
    assert raise_error
    assert isinstance(raise_error, bool)
    assert loader is None
    assert groups == ['gr']
    assert success_group is None
    assert failure_group is None
    assert py_dir is None
    assert parent is None


def test_pype_get_args_no_parent_context():
    """If args set use_parent_context should default False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'},
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args == {'a': 'b'}
    assert out is None
    assert not use_parent_context
    assert pipe_arg is None
    assert skip_parse
    assert raise_error
    assert not loader
    assert not groups
    assert not success_group
    assert not failure_group
    assert py_dir is None
    assert parent is None


def test_pype_get_pipeargs_no_skip_parse():
    """If pipeArgs set skipParse should default False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'a b c',
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args is None
    assert out is None
    assert not use_parent_context
    assert pipe_arg == ['a', 'b', 'c']
    assert not skip_parse
    assert raise_error
    assert not loader
    assert not groups
    assert not success_group
    assert not failure_group
    assert py_dir is None
    assert parent is None


def test_pype_get_args_and_pipearg():
    """Combine pipeArgs and args. Defaults useParentContext to False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'},
            'pipeArg': 'a b c',
        }
    })

    with get_arb_pipeline_scope(context):
        (pipeline_name,
         args,
         out,
         use_parent_context,
         pipe_arg,
         skip_parse,
         raise_error,
         loader,
         groups,
         success_group,
         failure_group,
         py_dir,
         parent) = pype.get_arguments(context)

    assert pipeline_name == 'pipe name'
    assert args == {'a': 'b'}
    assert out is None
    assert not use_parent_context
    assert pipe_arg == ['a', 'b', 'c']
    assert not skip_parse
    assert raise_error
    assert not loader
    assert not groups
    assert not success_group
    assert not failure_group
    assert py_dir is None
    assert parent is None

# region cascading


def get_scope_from_info(context, info):
    """Context must be in pipeline scope to get current pipe info."""
    pipeline = Pipeline('arb pipe')
    pipeline.pipeline_definition = PipelineDefinition(pipeline=None, info=info)
    return context.pipeline_scope(pipeline)


def test_pype_gets_args_loader_non_cascading():
    """Loader not cascading."""
    context = Context({
        'pype': {
            'name': 'pipe name'
        }
    })

    info = PipelineInfo(pipeline_name='arb',
                        loader='arbloader',
                        parent='noncascading/dir',
                        is_loader_cascading=False)

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader is None
    assert pype_args.parent is None


def test_pype_get_args_parent_cascading():
    """Same loader as parent with cascade."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'loader': 'arbloader'
        }
    })

    info = PipelineInfo(pipeline_name='arb',
                        loader='arbloader',
                        parent='cascading/parent')

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader == 'arbloader'
    assert pype_args.parent == 'cascading/parent'


def test_pype_get_args_parent_not_cascading_parent():
    """A non-cascading loader does not cascade parent."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'loader': 'arbloader'
        }
    })

    info = PipelineInfo(pipeline_name='arb',
                        loader='arbloader',
                        parent='noncascading/dir',
                        is_parent_cascading=False)

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader == 'arbloader'
    assert pype_args.parent is None


def test_pype_get_args_parent_cascading_different_loader_set():
    """Do not cascade when different loader than parent."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'loader': 'arbloader1'
        }
    })

    info = PipelineInfo(
        pipeline_name='arb',
        loader='arbloader2',
        parent='cascading/dir')

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader == 'arbloader1'
    assert pype_args.parent is None


def test_pype_get_args_parent_cascade_no_loader_set():
    """No loader set on parent and no loader on pype with cascade.

    This is very edge. pypyr core will normalize loader=None to the file
    loader, and to use a custom loader the dev would have to pass a non-None
    to run()/load_and_run() to begin with.

    It shouldn't ever happen short of some heavy customization where a custom
    client loader explicitly sets None the loader property on the Info object.
    Which, why? But just in case, here's a test.
    """
    context = Context({
        'pype': {
            'name': 'pipe name'
        }
    })

    info = PipelineInfo(
        pipeline_name='arb',
        loader=None,
        parent='cascading/parent')

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader is None
    assert pype_args.parent == 'cascading/parent'


def test_pype_get_args_parent_cascade_no_loader_no_parent_set():
    """No loader+parent set on parent & no loader+parent on pype with cascade.

    This is very edge. pypyr core will normalize loader=None to the file
    loader, and to use a custom loader the dev would have to pass a non-None
    to run()/load_and_run() to begin with.

    It shouldn't ever happen short of some heavy customization where a custom
    client loader explicitly sets None the loader property on the Info object.
    Which, why? But just in case, here's a test.
    """
    context = Context({
        'pype': {
            'name': 'pipe name'
        }
    })

    info = PipelineInfo(
        pipeline_name='arb',
        loader=None,
        parent=None)

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader is None
    assert pype_args.parent is None


def test_pype_get_parent_set_on_cascading_loader():
    """Always use parent when set even if cascading loader."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'loader': 'arbloader',
            'parent': 'arb/from/child'
        }
    })

    info = PipelineInfo(pipeline_name='arb',
                        loader='arbloader',
                        parent='ignore me')

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader == 'arbloader'
    assert pype_args.parent == 'arb/from/child'


def test_pype_resolve_from_parent_false_on_cascading_loader():
    """Ignore loader is_parent_cascading when resolveFromParent False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'loader': 'arbloader',
            'resolveFromParent': False
        }
    })

    info = PipelineInfo(pipeline_name='arb',
                        loader='arbloader',
                        parent='ignore me')

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader == 'arbloader'
    assert pype_args.parent is None


def test_pype_resolve_from_parent_true_on_cascading_loader():
    """Ignore loader is_parent_cascading when resolveFromParent False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'loader': 'arbloader',
            'resolveFromParent': True
        }
    })

    info = PipelineInfo(pipeline_name='arb',
                        loader='arbloader',
                        parent='from parent')

    with get_scope_from_info(context, info):
        pype_args = pype.get_arguments(context)

    assert pype_args.pipeline_name == 'pipe name'
    assert pype_args.loader == 'arbloader'
    assert pype_args.parent == 'from parent'

# endregion cascading

# endregion get_arguments

# region run_step


def mocked_run_pipeline(*args, **kwargs):
    """Check pipeline name set on context in child pipeline."""
    # assert (kwargs['name']
    #         == kwargs['context'].pipeline_name == 'pipe name')
    # assert (kwargs['pipeline_name'] == 'pipe name')
    pass


def test_pype_use_parent_context(mock_pipe):
    """Input pype use_parent_context True."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': True,
            'loader': 'test loader'
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader='test loader',
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once_with(context, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, using parent context.'),
        call('pyped pipe name.')]


def test_pype_use_parent_context_with_args(mock_pipe):
    """Input pype use_parent_context True with args."""
    context = Context({
        'k1': 'v1',
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'},
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': True,
            'loader': 'test loader'
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    # args merges into context
    merged_context = {
        'a': 'b',
        'k1': 'v1',
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'},
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': True,
            'loader': 'test loader'
        }
    }

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader='test loader',
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once_with(merged_context, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, using parent context.'),
        call('pyped pipe name.')]


def test_pype_no_parent_context(mock_pipe):
    """Input pype use_parent_context False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': False,
            'skipParse': True,
            'raiseError': True,
            'loader': 'test loader',
        }
    })
    # context.parent = 'arb/dir'
    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader='test loader',
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    # using empty/fresh new context
    mocked_runner.assert_called_once_with({}, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]


def test_pype_args(mock_pipe):
    """Input pype args used as context."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'}
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=None,
        parse_input=False,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    # using args as context
    mocked_runner.assert_called_once_with({'a': 'b'}, None)


def test_pype_args_with_shortcut(mock_pipe, monkeypatch):
    """Input pype args merged into shortcut used as context."""
    shortcuts = {'pipe name': {
        'pipeline_name': 'sc pipe',
        'args': {
            'a': 'og',
            'c': ['d']
        }
    }}

    monkeypatch.setattr('pypyr.config.config.shortcuts', shortcuts)

    context = Context({
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'}
        }
    })

    def l_and_r(context, parent):
        assert parent is None
        assert context == {'a': 'b', 'c': ['d']}
        context['c'].append('updated')
        context['e'] = 'new'

    mock_pipe.return_value.load_and_run_pipeline.side_effect = l_and_r

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='sc pipe',
        context_args=None,
        parse_input=False,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once()

    # original shortcut shouldn't mutate
    assert shortcuts == {'pipe name': {
        'pipeline_name': 'sc pipe',
        'args': {
            'a': 'og',
            'c': ['d']
        }
    }}

    # useParentContext is False, therfore no context mutations in parent.
    assert context == {
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'}
        }}


def test_pype_args_with_out(mock_pipe):
    """Input pype args used as context with out."""
    context = Context({
        'parentkey': 'parentvalue',
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'b'},
            'out': 'a'
        }
    })
    context.parent = 'arb/dir'
    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=None,
        parse_input=False,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    # using args as new context
    mocked_runner.assert_called_once_with({'a': 'b'}, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]

    assert context == {'parentkey': 'parentvalue',
                       'a': 'b',
                       'pype': {
                            'name': 'pipe name',
                            'args': {'a': 'b'},
                            'out': 'a'
                       }
                       }


def test_pype_args_with_mapping_out(mock_pipe):
    """Input pype args used as context with mapping out."""
    context = Context({
        'parentkey': 'parentvalue',
        'pype': {
            'name': 'pipe name',
            'args': {'a': 'av', 'b': 'bv', 'c': 'cv'},
            'out': {'new-a': 'a',
                    'new-c': 'c'}
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=None,
        parse_input=False,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once_with({'a': 'av', 'b': 'bv', 'c': 'cv'},
                                          None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]

    assert context == {'parentkey': 'parentvalue',
                       'new-a': 'av',
                       'new-c': 'cv',
                       'pype': {
                           'name': 'pipe name',
                           'args': {'a': 'av', 'b': 'bv', 'c': 'cv'},
                           'out': {'new-a': 'a',
                                   'new-c': 'c'}
                       }
                       }


def test_pype_no_skip_parse(mock_pipe):
    """Input pype use_parent_context False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': False,
            'skipParse': False,
            'raiseError': True
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=True,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once_with({}, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]


def test_pype_no_pipe_arg(mock_pipe):
    """Input pype use_parent_context False."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': None,
            'useParentContext': False,
            'skipParse': False,
            'raiseError': True,
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=None,
        parse_input=True,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once_with({}, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, without parent context.'),
        call('pyped pipe name.')]


def mocked_run_pipeline_with_runtime_error(*args, **kwargs):
    """Check pipeline name set on context in child pipeline with arb err."""
    assert (kwargs['pipeline_name']
            == kwargs['context'].pipeline_name == 'pipe name')
    assert (kwargs['pipeline_name'] == 'pipe name')
    raise RuntimeError('whoops')


def test_pype_use_parent_context_no_swallow(mock_pipe):
    """Input pype without swallowing error in child pipeline."""
    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.side_effect = RuntimeError('whoops')

    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': True
        }
    })

    with patch_logger('pypyr.steps.pype', logging.ERROR) as mock_logger_error:
        with pytest.raises(RuntimeError) as err_info:
            with get_arb_pipeline_scope(context):
                pype.run_step(context)

        assert str(err_info.value) == "whoops"

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner.assert_called_once_with({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': True
        }
    }, None)

    mock_logger_error.assert_called_once_with(
        'Something went wrong pyping pipe name. RuntimeError: whoops')


def test_pype_use_parent_context_with_swallow(mock_pipe):
    """Input pype swallowing error in child pipeline."""
    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.side_effect = RuntimeError('whoops')

    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': False,
            'loader': 'test loader'
        }
    })

    with patch_logger('pypyr.steps.pype', logging.ERROR) as mock_logger_error:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader='test loader',
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner.assert_called_once_with(context, None)

    mock_logger_error.assert_called_once_with(
        'Something went wrong pyping pipe name. RuntimeError: whoops')


def mocked_run_pipeline_with_stop(*args, **kwargs):
    """Check pipeline name set on context in child pipeline with Stop."""
    assert (kwargs['pipeline_name']
            == kwargs['context'].pipeline_name == 'pipe name')
    assert (kwargs['pipeline_name'] == 'pipe name')
    raise Stop()


def test_pype_use_parent_context_swallow_stop_error(mock_pipe):
    """Input pype doesn't swallow stop error in child pipeline."""
    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.side_effect = Stop()
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': False
        }
    })

    with patch_logger('pypyr.steps.pype', logging.ERROR) as mock_logger_error:
        with pytest.raises(Stop) as err_info:
            with get_arb_pipeline_scope(context):
                pype.run_step(context)

        assert isinstance(err_info.value, Stop)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader=None,
        groups=None,
        success_group=None,
        failure_group=None,
        py_dir=None
    )

    mocked_runner.assert_called_once_with(context, None)

    mock_logger_error.assert_not_called()


def test_pype_set_groups(mock_pipe):
    """Input pype with groups set."""
    context = Context({
        'pype': {
            'name': 'pipe name',
            'pipeArg': 'argument here',
            'useParentContext': True,
            'skipParse': True,
            'raiseError': True,
            'loader': 'test loader',
            'groups': 'testgroup',
            'success': 'successgroup',
            'failure': 'failuregroup',
            'pyDir': 'test dir'
        }
    })

    with patch_logger('pypyr.steps.pype', logging.INFO) as mock_logger_info:
        with get_arb_pipeline_scope(context):
            pype.run_step(context)

    mock_pipe.assert_called_once_with(
        name='pipe name',
        context_args=['argument', 'here'],
        parse_input=False,
        loader='test loader',
        groups=['testgroup'],
        success_group='successgroup',
        failure_group='failuregroup',
        py_dir='test dir'
    )

    mocked_runner = mock_pipe.return_value.load_and_run_pipeline
    mocked_runner.assert_called_once_with(context, None)

    assert mock_logger_info.mock_calls == [
        call('pyping pipe name, using parent context.'),
        call('pyped pipe name.')]
# endregion run_step


# region write_child_context_to_parent

def test_write_child_context_to_parent_wrong_type():
    """When out not a str, list or dict raise."""
    with pytest.raises(ContextError) as err_info:
        pype.write_child_context_to_parent(3, None, None)

    assert str(err_info.value) == (
        "pypyr.steps.pype pype.out should be a string, or a list or a dict. "
        "Instead, it's a <class 'int'>")


def test_write_child_context_to_parent_string():
    """Single string writes single key to parent."""
    parent = Context({'a': 'b'})
    child = Context({'c': 'd',
                     'e': 'f'})

    pype.write_child_context_to_parent('c', parent, child)

    assert parent == {'a': 'b',
                      'c': 'd'}


def test_write_child_context_to_parent_list():
    """Single string writes list of keys to parent."""
    parent = Context({'a': 'b'})
    child = Context({'c': 'd',
                     'e': 'f',
                    'g': 'h'})

    pype.write_child_context_to_parent(['c', 'g'], parent, child)

    assert parent == {'a': 'b',
                      'c': 'd',
                      'g': 'h'}


def test_write_child_context_to_parent_dict():
    """Single string maps keys to parent."""
    parent = Context({'a': 'b'})
    child = Context({'c': 'd',
                     'e': 'f',
                    'g': 'h'})

    pype.write_child_context_to_parent({'new-c': 'c',
                                        'new-g': 'g'},
                                       parent,
                                       child)

    assert parent == {'a': 'b',
                      'new-c': 'd',
                      'new-g': 'h'}


def test_write_child_context_to_parent_dict_with_formatting():
    """Single string maps keys to parent and formats child."""
    parent = Context({'a': 'b'})
    child = Context({'c': 'd',
                     'e': 'f',
                    'g': 'h and {e}'})

    pype.write_child_context_to_parent({'new-c': 'c',
                                        'new-g': 'g'},
                                       parent,
                                       child)

    assert parent == {'a': 'b',
                      'new-c': 'd',
                      'new-g': 'h and f'}
# endregion write_child_context_to_parent



"""pypyr pipeline runner.

This is the entrypoint for the pypyr API.

Use run() to run a pipeline.
"""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
import logging
from os import PathLike

from pypyr.context import Context
from pypyr.pipeline import Pipeline

logger = logging.getLogger(__name__)


def run(
    pipeline_name: str,
    args_in: list[str] | None = None,
    parse_args: bool | None = None,
    dict_in: dict | None = None,
    groups: list[str] | None = None,
    success_group: str | None = None,
    failure_group: str | None = None,
    loader: str | None = None,
    py_dir: str | bytes | PathLike | None = None
) -> Context:
    """Run a pipeline. pypyr's entrypoint.

    Call me if you want to run a pypyr pipeline from your own code.

    If you want to run a pipeline exactly like the cli does, use args_in to
    pass a list of str arguments for the pipeline's context_parser. If you
    already have a dict-like structure you want to use to initialize context,
    use dict_in instead. If you provide dict_in and no args_in, pypyr will
    assume you mean not to run the context_parser on the pipeline
    (parse_args=False) - if you do want to run the context_parser in this case,
    explicitly set parse_args=True.

    If you're invoking pypyr from your own application via the API, it's your
    responsibility to set up and configure logging. If you just want to
    replicate the log handlers & formatters that the pypyr cli uses, you can
    call pypyr.log.logger.set_root_logger() once and only once before invoking
    run() for every pipeline you want to run.

    Be aware that pypyr adds a NOTIFY - 25 custom log-level and notify()
    function to logging.

    {pipeline_name}.yaml should resolve from the current working directory if
    you are using the default file loader.

    You only need to specify py_dir if your pipeline relies on custom modules
    that are NOT installed in the current Python environment. For convenience,
    pypyr allows pipeline authors to use ad hoc python modules that are not
    installed in the current environment by looking for these in py_dir 1st.

    Regardless of whether you set py_dir or not, be aware that if you are using
    the default file loader, pypyr will also add the pipeline's immediate
    parent directory to sys.path (only if it's not been added already), so that
    each pipeline can reference ad hoc modules relative to itself in the
    filesystem.

    Therefore you do NOT need to set py_dir if your ad hoc custom modules are
    relative to the pipeline itself.

    If your pipelines are only using built-in functionality, you don't need to
    set py_dir.

    If pipeline_name found in config.shortcuts, will use the configured values
    from the shortcut to find & run the pipeline. If a shortcut does not have
    a value for a particular argument, will fallback to the run() function's
    inputs, if any. If you pass args_in and/or dict_in and a shortcut is found
    for pipeline_name, will merge the function's input arguments into the
    shortcut's pipe_arg and args respectively.

    Example: Run ./dir/pipe-name.yaml, resolve ad hoc custom modules from the
    current directory and initialize context with dict {'a': 'b'}:

    context = run('dir/pipe-name', dict_in={'a': 'b'}, py_dir=Path.cwd())

    Args:
        pipeline_name (str): Name of pipeline, sans .yaml at end.
        args_in (list[str]): All the input arguments after the pipeline name
            from cli.
        parse_args (bool): run context_parser in pipeline. Default True.
        dict_in (dict): Dict-like object to initialize the Context.
        groups: (list[str]): Step-group names to run in pipeline.
            Default is ['steps'].
        success_group (str): Step-group name to run on success completion.
            Default is on_success.
        failure_group: (str): Step-group name to run on pipeline failure.
            Default is on_failure.
        loader (str): optional. Absolute name of pipeline loader module.
            If not specified will use pypyr.loaders.file.
        py_dir (Path-like): Custom python modules resolve from this dir.

    Returns:
        pypyr.context.Context(): The pypyr context as it is after the pipeline
                                  completes.
    """
    logger.debug("starting pypyr")

    pipeline, args = Pipeline.new_pipe_and_args(name=pipeline_name,
                                                context_args=args_in,
                                                parse_input=parse_args,
                                                dict_in=dict_in,
                                                loader=loader,
                                                groups=groups,
                                                success_group=success_group,
                                                failure_group=failure_group,
                                                py_dir=py_dir)

    context = Context(args) if args else Context()

    pipeline.run(context)

    logger.debug("pypyr done")

    return context



"""types.py unit tests."""
import pypyr.utils.types as types

# region are_all_this_type


def test_all_objects_of_type():
    """All of type should be True."""
    assert types.are_all_this_type(int, 1, 2, 3, 4)


def test_none_objects_of_type():
    """None of type should be False."""
    assert not types.are_all_this_type(str, 1, 2, 3, 4)


def test_one_obj_of_type():
    """Only 1 input of type should be True."""
    assert types.are_all_this_type(bool, False)


def test_some_objects_of_type():
    """Odd one out should be False."""
    assert not types.are_all_this_type(str, 1, '2', 3, 4)

# endregion are_all_this_type

# region cast_to_bool


def test_cast_to_bool():
    """Special strings eval true."""
    assert types.cast_to_bool('TRUE') is True
    assert types.cast_to_bool('true') is True
    assert types.cast_to_bool(1) is True
    assert types.cast_to_bool(1.0) is True
    assert types.cast_to_bool(2) is True

    assert types.cast_to_bool(0) is False
    assert types.cast_to_bool('FALSE') is False
    assert types.cast_to_bool('arb') is False
    assert types.cast_to_bool(None) is False
# endregion cast_to_bool

# region cast_to_type


def test_cast_to_type_with_conversion():
    """In obj casts to out type."""
    out = types.cast_to_type(12, str)
    assert isinstance(out, str)
    assert out == "12"


def test_cast_to_type_not_needed():
    """In obj is already out type, no conversion needed."""
    out = types.cast_to_type(12, int)
    assert isinstance(out, int)
    assert out == 12


def test_cast_to_type_str_to_bool():
    """String to bool always true."""
    out = types.cast_to_type('False', bool)
    assert isinstance(out, bool)
    assert out
# endregion cast_to_type

# region empty_if_none


def test_empty_if_none():
    """Empty if none returns empty string when None."""
    assert types.empty_if_none(None) == ''
    assert types.empty_if_none('') == ''
    assert types.empty_if_none('arb') == 'arb'
    assert types.empty_if_none(True) is True
    v = ValueError()
    assert types.empty_if_none(v) == v
# endregion empty_if_none



"""Unit tests for pypyr.steps.venv.

Most of the testing happens in pypyr.steps.dsl.venv & pypyr.venv.

All this needs to test is that the step calls through to the underlying core.
"""
from pathlib import Path
from types import SimpleNamespace
from unittest.mock import patch

from pypyr.context import Context
from pypyr.steps import venv


def get_simple_context():
    """Create a test SimpleNamespace to serve as EnvBuilder context."""
    context = SimpleNamespace()
    context.env_exec_cmd = '/python'
    context.bin_path = '/venv/bin'
    context.env_dir = '/venv'
    return context


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_create(mock_builder):
    """Create a venv calls through to the venv dsl."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    venv.run_step(Context({'venv': '/arb'}))

    expected_path = str(Path('/arb').expanduser().resolve())

    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_called_once_with(context)
    mocked_builder.pip_install_extras.assert_not_called()



"""contextclear.py unit tests."""
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.contextclear
import pytest


def test_context_clear_throws_on_empty_context():
    """Context must exist."""
    with pytest.raises(KeyNotInContextError):
        pypyr.steps.contextclear.run_step(Context())


def test_context_clear_throws_on_contextset_missing():
    """Context Clear must exist in context."""
    with pytest.raises(KeyNotInContextError):
        pypyr.steps.contextclear.run_step(Context({'arbkey': 'arbvalue'}))


def test_context_clear_pass():
    """Context clear success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'key4': 'value4',
        'contextClear': [
            'key2',
            'key4',
            'contextClear'
        ]
    })

    pypyr.steps.contextclear.run_step(context)

    assert len(context) == 2
    assert context['key1'] == 'value1'
    assert 'key2' not in context
    assert context['key3'] == 'value3'
    assert 'key4' not in context
    # recursion ahoy
    assert 'contextClear' not in context


def test_context_clear_no_raise_if_keys_dont_exist():
    """Context clear passes even if keys not in context."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'key4': 'value4',
        'contextClear': [
            'key2',
            'key4',
            'key5',
            'key1',
        ]
    })

    pypyr.steps.contextclear.run_step(context)

    assert len(context) == 2
    assert 'key1' not in context
    assert 'key2' not in context
    assert context['key3'] == 'value3'
    assert 'key4' not in context
    # recursion ahoy
    assert context['contextClear'] == [
        'key2',
        'key4',
        'key5',
        'key1',
    ]



"""pypyr/steps/set.py unit tests."""
import pytest

from pypyr.context import Context
from pypyr.dsl import PyString
from pypyr.errors import KeyNotInContextError
import pypyr.steps.set


def test_set_throws_on_empty_context():
    """Context must exist."""
    with pytest.raises(KeyNotInContextError):
        pypyr.steps.set.run_step(Context())


def test_set_throws_on_contextset_missing():
    """Input set must exist in context."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.set.run_step(Context({'arbkey': 'arbvalue'}))

    assert str(err_info.value) == ("context['set'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.set.")


def test_set_pass_no_substitutions():
    """Input set success case with no substitutions."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'set': {
            'key2': 'value4',
            'key4': 'value5'
        }
    })

    pypyr.steps.set.run_step(context)

    assert context['key1'] == 'value1'
    assert context['key2'] == 'value4'
    assert context['key3'] == 'value3'
    assert context['key4'] == 'value5'


def test_set_pass_substitutions():
    """Input set success case with substitutions."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'set': {
            'key2': '{key1}',
            'key4': '{key3}'
        }
    })

    pypyr.steps.set.run_step(context)

    assert context['key1'] == 'value1'
    assert context['key2'] == 'value1'
    assert context['key3'] == 'value3'
    assert context['key4'] == 'value3'


def test_set_pass_different_types():
    """Input set success case with substitutions of non strings."""
    context = Context({
        'k1': 33,
        'k2': 123.45,
        'k3': False,
        'set': {
            'kint': '{k1}',
            'kfloat': '{k2}',
            'kbool': '{k3}'
        }
    })

    pypyr.steps.set.run_step(context)

    assert context['kint'] == 33
    assert context['k1'] == 33
    assert context['kfloat'] == 123.45
    assert context['k2'] == 123.45
    assert not context['kbool']
    assert isinstance(context['kbool'], bool)
    assert not context['k3']


def test_set_list():
    """Simple list."""
    context = Context({'ctx1': 'ctxvalue1',
                       'ctx2': 'ctxvalue2',
                       'ctx3': 'ctxvalue3',
                       'set': {
                           'output': ['k1', 'k2', '{ctx3}', True, False, 44]
                       }})

    pypyr.steps.set.run_step(context)

    output = context['output']
    assert output[0] == 'k1'
    assert output[1] == 'k2'
    assert output[2] == 'ctxvalue3'
    assert output[3]
    assert not output[4]
    assert output[5] == 44


def test_set_tuple():
    """Simple tuple."""
    context = Context({'ctx1': 'ctxvalue1',
                       'ctx2': 'ctxvalue2',
                       'ctx3': 'ctxvalue3',
                       'set': {
                           'output': ('k1', 'k2', '{ctx3}', True, False, 44)
                       }})

    pypyr.steps.set.run_step(context)

    output = context['output']
    assert output[0] == 'k1'
    assert output[1] == 'k2'
    assert output[2] == 'ctxvalue3'
    assert output[3]
    assert not output[4]
    assert output[5] == 44


def test_set_set():
    """Simple set."""
    input_obj = {'k1', 'k2', '{ctx3}', True, False, 44}
    context = Context({'ctx1': 'ctxvalue1',
                       'ctx2': 'ctxvalue2',
                       'ctx3': 'ctxvalue3',
                       'set': {
                           'output': input_obj
                       }})

    pypyr.steps.set.run_step(context)

    output = context['output']
    assert len(output) == len(input_obj)
    diffs = output - input_obj
    assert len(diffs) == 1
    assert 'ctxvalue3' in diffs


def test_set_nested():
    """Straight deepish copy with no formatting."""
    # dict containing dict, list, dict-list-dict, tuple, dict-tuple-list
    input_obj = {'k1': 'v1',
                 'k2': 'v2',
                 'k3': 'v3',
                 'k4': [
                     1,
                     2,
                     '3here',
                     {'key4.1': 'value4.1',
                      'key4.2': 'value4.2',
                      'key4.3': {
                          '4.3.1': '4.3.1value',
                          '4.3.2': '4.3.2value'}}
                 ],
                 'k5': {'key5.1': 'value5.1', 'key5.2': 'value5.2'},
                 'k6': ('six6.1', False, [0, 1, 2], 77, 'sixend'),
                 'k7': 'simple string to close 7'
                 }

    context = Context({'ctx1': 'ctxvalue1',
                       'ctx2': 'ctxvalue2',
                       'ctx3': 'ctxvalue3',
                       'set': {
                           'output': input_obj}})

    pypyr.steps.set.run_step(context)

    output = context['output']
    assert output == input_obj
    assert output is not context
    # verify this was a deep copy - obj refs has to be different for nested
    assert id(output['k4']) != id(input_obj['k4'])
    assert id(output['k4'][3]['key4.3']) != id(input_obj['k4'][3]['key4.3'])
    assert id(output['k5']) != id(input_obj['k5'])
    assert id(output['k6']) != id(input_obj['k6'])
    assert id(output['k6'][2]) != id(input_obj['k6'][2])
    assert id(output['k7']) == id(input_obj['k7'])

    # and proving the theory: mutating output does not touch input
    assert output['k4'][1] == 2
    output['k4'][1] = 88
    assert input_obj['k4'][1] == 2
    assert output['k4'][1] == 88


def test_get_formatted_iterable_nested_with_formatting():
    """Straight deepish copy with formatting."""
    # dict containing dict, list, dict-list-dict, tuple, dict-tuple-list, bytes
    input_obj = {'k1': 'v1',
                 'k2': 'v2_{ctx1}',
                 'k3': bytes('v3{ctx1}', encoding='utf-8'),
                 'k4': [
                     1,
                     2,
                     '3_{ctx4}here',
                     {'key4.1': 'value4.1',
                      '{ctx2}_key4.2': 'value_{ctx3}_4.2',
                      'key4.3': {
                          '4.3.1': '4.3.1value',
                          '4.3.2': '4.3.2_{ctx1}_value'}}
                 ],
                 'k5': {'key5.1': 'value5.1', 'key5.2': 'value5.2'},
                 'k6': ('six6.1', False, [0, 1, 2], 77, 'six_{ctx1}_end'),
                 'k7': 'simple string to close 7'
                 }

    context = Context(
        {'ctx1': 'ctxvalue1',
         'ctx2': 'ctxvalue2',
         'ctx3': 'ctxvalue3',
         'ctx4': 'ctxvalue4',
         'set': {
             'output': input_obj,
             '{ctx1}': 'substituted key'}})

    pypyr.steps.set.run_step(context)

    output = context['output']
    assert output != input_obj

    # verify formatted strings
    assert input_obj['k2'] == 'v2_{ctx1}'
    assert output['k2'] == 'v2_ctxvalue1'

    assert input_obj['k3'] == b'v3{ctx1}'
    assert output['k3'] == b'v3{ctx1}'

    assert input_obj['k4'][2] == '3_{ctx4}here'
    assert output['k4'][2] == '3_ctxvalue4here'

    assert input_obj['k4'][3]['{ctx2}_key4.2'] == 'value_{ctx3}_4.2'
    assert output['k4'][3]['ctxvalue2_key4.2'] == 'value_ctxvalue3_4.2'

    assert input_obj['k4'][3]['key4.3']['4.3.2'] == '4.3.2_{ctx1}_value'
    assert output['k4'][3]['key4.3']['4.3.2'] == '4.3.2_ctxvalue1_value'

    assert input_obj['k6'][4] == 'six_{ctx1}_end'
    assert output['k6'][4] == 'six_ctxvalue1_end'

    # verify this was a deep copy - obj refs has to be different for nested
    assert id(output['k4']) != id(input_obj['k4'])
    assert id(output['k4'][3]['key4.3']) != id(input_obj['k4'][3]['key4.3'])
    assert id(output['k5']) != id(input_obj['k5'])
    assert id(output['k6']) != id(input_obj['k6'])
    assert id(output['k6'][2]) != id(input_obj['k6'][2])
    # short strings are interned in python, so id is the same
    assert id(output['k7']) == id(input_obj['k7'])
    output['k7'] = 'mutate 7 on new'
    assert input_obj['k7'] == 'simple string to close 7'
    assert output['k7'] == 'mutate 7 on new'

    assert context['ctxvalue1'] == 'substituted key'


def test_no_clash_with_builtin_set():
    """Module name doesn't clash with built-in set()."""
    context = Context(
        {'set': {
            'direct': set([1, 2]),
            'pystring': PyString('set([3, 4])')
        }
        })

    pypyr.steps.set.run_step(context)

    assert len(context) == 2
    assert context['direct'] == {1, 2}
    assert context['pystring'] == {3, 4}



"""Tests for pypyr.steps.cmd that touch the filesystem.

Can't use pyfakefs for these because the subprocess lib uses C libs to hit the
fs rather than the patch-able Python files access.
"""
from pathlib import Path
import subprocess
import tempfile

import pytest

from pypyr.config import config
from pypyr.context import Context
import pypyr.steps.cmd

is_windows = config.is_windows


def get_cmd(posix, win):
    """Return posix or win depending on current platform."""
    return win if is_windows else posix


@pytest.fixture
def temp_dir():
    """Make tmp dir in testfiles/out. Yields pathlib.Path."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield Path(tmp_dir)


def test_cmd_run_has_list_input_output_to_file(temp_dir):
    """List input to run complex dict save all output to files."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('echo two three',
                   r'tests\testfiles\cmds\echo.bat "two three"')

    stdout = temp_dir.joinpath('cmdout/stdout')
    stderr = temp_dir.joinpath('cmdout2/stderr')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2],
            'stdout': stdout,
            'stderr': stderr}
    })
    pypyr.steps.cmd.run_step(context)

    assert 'cmdOut' not in context

    assert stdout.read_text() == 'one\ntwo three\n'
    assert stderr.read_text() == ''


def test_cmd_run_has_list_input_error_throws_and_stderr(temp_dir):
    """Write to stdout and stderr."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                   r'tests\testfiles\cmds\exitwitherr.bat')

    # won't run, because 2 errors. test assert checks it doesn't run.
    cmd3 = get_cmd('echo three',
                   r'tests\testfiles\cmds\echo.bat three')

    stdout = temp_dir.joinpath('stdout')
    stderr = temp_dir.joinpath('stderr')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2, cmd3],
            'stdout': stdout,
            'stderr': stderr}
    })

    with pytest.raises(subprocess.CalledProcessError):
        pypyr.steps.cmd.run_step(context)

    assert stdout.read_text() == 'one\n'
    assert stderr.read_text() == 'arb err here\n'


def test_cmd_run_has_list_input_error_stderr_to_stdout(temp_dir):
    """Redirect stderr to stdout."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                   r'tests\testfiles\cmds\exitwitherr.bat')

    # won't run, because 2 errors. test assert checks it doesn't run.
    cmd3 = get_cmd('echo three',
                   r'tests\testfiles\cmds\echo.bat three')

    stdout = temp_dir.joinpath('stdout')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2, cmd3],
            'stdout': stdout,
            'stderr': '/dev/stdout'}
    })

    with pytest.raises(subprocess.CalledProcessError):
        pypyr.steps.cmd.run_step(context)

    assert stdout.read_text() == 'one\narb err here\n'


def test_cmd_run_has_list_input_overwrite_vs_append(temp_dir):
    """Write to stdout and stderr with append true/false, default overwrite."""
    stdout = temp_dir.joinpath('stdout')
    stderr = temp_dir.joinpath('stderr')

    cmd1 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh one',
                   r'tests\testfiles\cmds\echo-out-and-err.bat one')

    cmd2 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh two',
                   r'tests\testfiles\cmds\echo-out-and-err.bat two')

    context = Context({
        'cmd': {
            'run': [cmd1, cmd2],
            'stdout': stdout,
            'stderr': stderr}
    })

    pypyr.steps.cmd.run_step(context)

    assert stdout.read_text() == 'stdout one\nstdout two\n'
    assert stderr.read_text() == 'stderr one\nstderr two\n'

    # now overwrite these by default on 2nd invocation
    cmd3 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh three',
                   r'tests\testfiles\cmds\echo-out-and-err.bat three')

    cmd4 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh four',
                   r'tests\testfiles\cmds\echo-out-and-err.bat four')

    context = Context({
        'cmd': {
            'run': [cmd3, cmd4],
            'stdout': stdout,
            'stderr': stderr}
    })

    pypyr.steps.cmd.run_step(context)

    assert stdout.read_text() == 'stdout three\nstdout four\n'
    assert stderr.read_text() == 'stderr three\nstderr four\n'

    # now append
    cmd5 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh five',
                   r'tests\testfiles\cmds\echo-out-and-err.bat five')

    cmd6 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh six',
                   r'tests\testfiles\cmds\echo-out-and-err.bat six')

    context = Context({
        'cmd': {
            'run': [cmd5, cmd6],
            'stdout': stdout,
            'stderr': stderr,
            'append': True}
    })

    pypyr.steps.cmd.run_step(context)

    assert stdout.read_text() == (
        'stdout three\nstdout four\nstdout five\nstdout six\n')
    assert stderr.read_text() == (
        'stderr three\nstderr four\nstderr five\nstderr six\n')



"""Custom exceptions for pypyr.

All pypyr specific exceptions derive from Error.

Do NOT import any other pypyr modules here. This is to prevent circular
imports. Every other module in pypyr potentially uses this module.
"""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Iterator, Sequence
import signal


def get_error_name(error):
    """Return canonical error name as string.

    For builtin errors like ValueError or Exception, will return the bare
    name, like ValueError or Exception.

    For all other exceptions, will return modulename.errorname, such as
    arbpackage.mod.myerror

    Args:
        error: Exception object.

    Returns:
        str. Canonical error name.

    """
    error_type = type(error)
    if error_type.__module__ in ['__main__', 'builtins']:
        return error_type.__name__
    else:
        return f'{error_type.__module__}.{error_type.__name__}'


class Error(Exception):
    """Base class for all pypyr exceptions."""


class ConfigError(Error):
    """Error loading configuration."""


class ContextError(Error):
    """Error in the pypyr context."""


class HandledError(Error):
    """Error that has already been saved to errors context collection."""


class KeyInContextHasNoValueError(ContextError):
    """pypyr context[key] doesn't have a value."""


class KeyNotInContextError(ContextError, KeyError):
    """Key not found in the pypyr context."""

    def __str__(self):
        """Avoid KeyError custom error formatting."""
        return super(Exception, self).__str__()


class LoopMaxExhaustedError(Error):
    """Max attempts reached during looping."""


class MultiError(Error):
    """Aggregate exception."""

    def __init__(self, message: str | None = None,
                 errors: list[Exception] | None = None):
        """Create a MultiError exception."""
        super().__init__(message, errors)
        self.message = message
        self.errors: list[Exception] = errors if errors else []

    def append(self, error: Exception) -> None:
        """Add exception to the MultiError."""
        self.errors.append(error)

    def extend(self, errors: Sequence) -> None:
        """Extend input errors into MultiError list."""
        self.errors.extend(errors)

    @property
    def has_errors(self) -> bool:
        """Return True if contains exceptions."""
        return len(self.errors) > 0

    def __getitem__(self, index: int) -> Exception:
        """Get exception by index."""
        return self.errors[index]

    def __iter__(self) -> Iterator[Exception]:
        """Iterate through the exceptions in the MultiError."""
        return iter(self.errors)

    def __len__(self) -> int:
        """Get number of errors contained in the MultiError."""
        return len(self.errors)

    def __repr__(self) -> str:
        """Get string repr."""
        args = []
        if self.message:
            args.append(f'message={repr(self.message)}')

        if self.errors:
            args.append(f'errors={repr(self.errors)}')

        args_str = ', '.join(args)
        return f"{type(self).__name__}({args_str})"

    def __str__(self) -> str:
        """Get human friendly string."""
        msg = f'{self.message}' if self.message else (
            'Multiple error(s) occurred:' if len(self.errors) > 1
            else 'An error occurred:')
        errors = "\n".join([f'{type(e).__name__}: {e}' for e in self.errors])
        return f"{msg}\n{errors}" if errors else msg


class PipelineDefinitionError(Error):
    """Pipeline definition incorrect. Likely a yaml error."""


class PipelineNotFoundError(Error):
    """Pipeline not found in working dir or in pypyr install dir."""


class PlugInError(Error):
    """Pypyr plug - ins should sub - class this."""


class PyModuleNotFoundError(Error, ModuleNotFoundError):
    """Could not load python module because it wasn't found."""


class SubprocessError(Error):
    """Error on executable or shell run as a sub-process.

    Attributes:
        cmd (str | bytes | Path): The cmd or shell instruction that caused
            the error.
        returncode (int): Return code from sub-process. 0 usually means OK.
        stdout (str | bytes | None): stdout output, if any.
        stderr (str | bytes | None): stderr output, if any.
    """

    def __init__(self, returncode, cmd, stdout=None, stderr=None) -> None:
        """Initialize class."""
        super().__init__(returncode, cmd, stdout, stderr)
        self.returncode = returncode
        self.cmd = cmd
        self.stdout = stdout
        self.stderr = stderr

    def __repr__(self) -> str:
        """Convert to repr."""
        args = ['returncode={!r}'.format(self.returncode),
                'cmd={!r}'.format(self.cmd)]
        if self.stdout is not None:
            args.append('stdout={!r}'.format(self.stdout))
        if self.stderr is not None:
            args.append('stderr={!r}'.format(self.stderr))
        return "{}({})".format(type(self).__name__, ', '.join(args))

    def __str__(self) -> str:
        """Convert to string."""
        if self.returncode and self.returncode < 0:
            try:
                sig = repr(signal.Signals(-self.returncode))
                return f"Command '{self.cmd}' died with {sig}."
            except ValueError:
                return (f"Command '{self.cmd}' died with unknown signal "
                        + f"{-self.returncode}.")
        else:
            stderr = f" stderr: {self.stderr}" if self.stderr else ''
            return (f"Command '{self.cmd}' returned non-zero exit status "
                    + f"{self.returncode}.{stderr}")

# region Control of Flow Instructions


class Stop(Error):
    """Control of flow. Stop all execution."""


class StopPipeline(Stop):
    """Control of flow. Stop current pipeline execution."""


class StopStepGroup(Stop):
    """Control of flow. Stop current step - group execution."""


class ControlOfFlowInstruction(Error):
    """Control of flow instructions should inherit from this.

    Attributes:
        groups: list of str. List of step - groups to execute.
        success_group: str. Step - group to execute on success condition.
        failure_group: str. Step - group to execute on failure condition.
        original_config: tuple. (key, value). Key-Value pair of context dict
                        item used as input on the step that created this
                        ControlOfFlowInstruction.
    """

    def __init__(self, groups, success_group, failure_group, original_config):
        """Initialize the control of flow instruction.

        Args:
            groups: list of str. List of step - groups to execute.
            success_group: str. Step - group to execute on success condition.
            failure_group: str. Step - group to execute on failure condition.
            original_config: tuple. (key, value). Key/Value pair of Context
                             input for the control-of-flow step.
        """
        self.groups = groups
        self.success_group = success_group
        self.failure_group = failure_group
        self.original_config = original_config


class Call(ControlOfFlowInstruction):
    """Stop current step, call a new step group, resume current step after."""


class Jump(ControlOfFlowInstruction):
    """Stop step execution and jump to a new step group."""

# endregion Control of Flow Instructions



"""pypyr dynamic modules, namespaces and path discovery.

Load modules dynamically, find things on file-system.
"""
import ast
import builtins
from collections import ChainMap
import importlib
import logging
from pathlib import Path
import sys
from threading import Lock

from pypyr.errors import PyModuleNotFoundError

logger = logging.getLogger(__name__)


class _ChainMapPretendDict(ChainMap, dict):  # type: ignore
    # mypy gets confused with multiple inheritance, thus ignore
    """It's a ChainMap. But it will pass an isinstance check for dict.

    Don't Use Me. Seriously.

    The only reason for this is to use as globals with eval/exec. Because
    eval/exec only accepts dict type as globals argument.

    However, for the purposes of pypyr, rather than make a whole copy of the
    context dict each and every time a py expression evaluates, it's much
    more performance efficient to use a chainmap to chain together the dynamic
    code's global namespace and context.

    That does not mean what you see here is a good idea. It's not. It's the
    least bad of other bad options.

    Don't pickle me. Re-instantiate instead.

    There are 2 storage mechanisms here:
    - the dict instance itself
    - the ChainMap maps attribute (a list of dicts, basically.)

    The ChainMap overrides most of the accessors you'd associate with dict
    (get, set, items, etc.) because it is 1st in the mro.

    See docs/adr/0001-namespace-on-eval-and-exec.md
    """

    __slots__ = ()

    def __init__(self, *maps):
        """Initialize the dict to builtins, and the chainmap to *maps.

        Args:
            maps: mapping dict-like objects to add to the chainmap.
        """
        dict.__setitem__(self, '__builtins__', builtins.__dict__)
        super().__init__(*maps)


class ImportVisitor(ast.NodeVisitor):
    """Parse python import and import from syntax.

    Use this to parse python syntax of import/import from, import the modules
    and save the result to the imported_namespace namespace dictionary.

    Only supports absolute imports, not relative. Does not support wildcard
    style 'from x import *'. But you weren't planning on doing *that* anyway,
    I hope.

    Supports source like:
        import x
        import x as y
        import x.y
        import x.y as z
        from x import y
        from x import y as z
        from x import y, z
        from a.b import c as d, e as f

    Usage example:
        ImportVisitor().visit(ast.parse('from mod.sub import attr as alias'))

    Will result in:
        imported_namespace == {'alias': <<attr in mod.sub module>>}

    Attributes:
        imported_namespace (dict): Namespace dictionary of imported references.
    """

    def __init__(self):
        """Initialize me."""
        self.imported_namespace = {}

    def get_namespace(self, source):
        """Parse source, import modules & return namespace.

        You might as well call this instead of visit().

        Args:
            source (str): String of Python source code.

        Return:
            Namespace dictionary of imported references.
        """
        self.visit(ast.parse(source))
        return self.imported_namespace

    def _set_namespace(self, alias, obj):
        """Add imported object to namespace dictionary.

        Args:
            alias (ast.alias): Use asname for alias if it exists, else fall
                back to name.
            obj (any): Imported module or attribute like class or function.

        Returns: None
        """
        as_name = alias.asname if alias.asname else alias.name
        self.imported_namespace[as_name] = obj

    def visit_Import(self, node):
        """Process syntax nodes of form: import module."""
        for alias in node.names:
            if alias.asname:
                imported_module = importlib.import_module(alias.name)
            else:
                # if no alias, 'import mod.sub' has to bind 'mod' to the
                # imported parent obj.
                parent, dot, _ = alias.name.partition('.')
                if dot:
                    # mod.sub1.sub2 should save to namespace as parent 'mod'
                    alias.asname = parent
                    # __import__(), although discouraged, is how to get the
                    # top-level module - this is the obj that is bound by the
                    # name in the namespace
                    imported_module = __import__(alias.name)
                else:
                    imported_module = importlib.import_module(alias.name)

            self._set_namespace(alias, imported_module)

    def visit_ImportFrom(self, node):
        """Process syntax nodes of form: from parent import child."""
        if node.level > 0:
            raise TypeError("you can't use relative imports here. "
                            "use absolute imports instead.")
        imported_module = importlib.import_module(node.module)
        for alias in node.names:
            try:
                imported_obj = getattr(imported_module, alias.name)
            except AttributeError:
                # if no attribute, might be form: from mod import submod
                imported_obj = importlib.import_module(
                    f'{node.module}.{alias.name}')

            self._set_namespace(alias, imported_obj)


_sys_path_lock = Lock()
_known_dirs = set()


def add_sys_path(path):
    """Add path to sys.path if it doesn't exist in sys.path already.

    Only add paths that actually exist.

    Do this under a shared lock to prevent duplicates in sys.path.

    _known_dirs does not mean a path exists, it means the logic around
    whether to add a path or not has run.

    Args:
        path (Path-like): path to add to sys.path
    """
    if path in _known_dirs:
        return

    path_obj = path if isinstance(path, Path) else Path(path)

    if not path_obj.exists():
        _known_dirs.add(path)
        return

    # sys path doesn't accept Path
    path_str = str(path_obj)  # .resolve(True)? instead for extended paths?
    logger.debug("adding %s to sys.paths", path_str)
    with _sys_path_lock:
        if path_str not in sys.path:
            # append not insert - don't overwrite user's prior sys.path
            sys.path.append(path_str)

    _known_dirs.add(path)


def get_module(module_abs_import):
    """Use importlib to get the module dynamically.

    Get instance of the module specified by the module_abs_import.
    This means that module_abs_import must be resolvable from this package.

    Args:
        module_abs_import: string. Absolute name of module to import.

    Raises:
        PyModuleNotFoundError: if module not found.

    """
    logger.debug("starting")
    logger.debug("loading module %s", module_abs_import)
    try:
        imported_module = importlib.import_module(module_abs_import)
        logger.debug("done")
        return imported_module
    except ModuleNotFoundError as err:
        if err.name != module_abs_import:
            extended_msg = (
                f'error importing module {err.name} in {module_abs_import}')
            logger.error("Couldn't import module %s in this module: %s",
                         err.name,
                         module_abs_import)
        else:
            extended_msg = (
                f"{module_abs_import}.py should be in your pipeline dir, or "
                "in your working dir, or it should be installed in the "
                "current python env."
                "\nIf you have 'package.sub.mod' your pipeline dir should "
                "contain ./package/sub/mod.py\n"
                "If you specified 'mymodulename', your pipeline dir should "
                "contain ./mymodulename.py\n"
                "If the module is not in your pipeline dir nor in the current "
                "working dir, it must exist in your current python env - so "
                "you should have run pip install or setup.py")
            logger.error("The module doesn't exist. "
                         "Looking for a file like this: %s",
                         module_abs_import)
        raise PyModuleNotFoundError(extended_msg) from err



"""Arbitrary callables for testing."""


class ArbCallable():
    """Arb callable test class."""

    def __init__(self, ctorarg):
        """Arbitrary ctor."""
        self.ctorarg = ctorarg

    def __call__(self, arg1):
        """Make it callable."""
        return f'from callable: {self.ctorarg} {arg1}'



"""pypyr step that runs sub-processes asynchronously.

This runs an executable, it does not invoke the shell. Therefore you cannot
use things like exit, return, shell pipes, filename wildcards,
environment variable expansion, and expansion of ~ to a user’s home
directory. See pypyr.steps.shells for shell invocation.
"""
import logging
from pypyr.steps.dsl.cmdasync import AsyncCmdStep

logger = logging.getLogger(__name__)


def run_step(context):
    """Run commands, programs or executables asynchronously, in parallel.

    Context is a pypyr.context.Context. This is dict-like.

    Context must contain the following keys:
        cmds:
            - <<cmd string 1>>
            - <<cmd string 2>>

    where <<cmd string>> is command + args to execute.

    You can alternatively use the expanded syntax to override default options:

    cmds:
        run: list[str | list[str]]. mandatory. command + args to execute.
            If list entry is another list[str], the sub-list will run in
            serial.
        save: bool. defaults False. save output to cmdOut. Treats output
            as text in the system's encoding and removes newlines at end.
        cwd: str/Pathlike. optional. Working directory for these commands.
        bytes (bool): Default False. When `save` return output bytes from
            cmds unaltered, without applying any encoding & text newline
            processing.
        encoding (str): Default None. When `save`, decode output with
            this encoding. The default of None uses the system encoding and
            should "just work".
        stdout (str | Path): Default None. Write stdout to this file path.
            Special value `/dev/null` writes to the system null device.
        stderr (str | Path): Default None. Write stderr to this file path.
            Special value `/dev/null` writes to the system null device.
            Special value `/dev/stdout` redirects err output to stdout.
        append (bool): Default False. When stdout/stderr a file, append
            rather than overwrite. Default is to overwrite.

    In expanded syntax, `run` can be a simple string or a list:
        cmds:
          run:
            - ./my-executable --arg
            - ./another-executable --arg
          save: False
          cwd: ./path/here

    Will execute the command string in as a sub-process.
    Escape curly braces: if you want a literal curly brace, double it like
    {{ or }}.

    If save is True, will save the output as a list of results in the order
    executed. Each result is a pypyr.subproc.SubprocessResult, with the schema:
    cmd: the cmd/args executed
    returncode: 0
    stdout: 'stdout str here. None if empty.'
    stderr: 'stderr str here. None if empty.'

    cmdOut.returncode is the exit status of the called process. Typically 0
    means OK. A negative value -N indicates that the child was terminated by
    signal N (POSIX only).

    If the sub-process couldn't spawn at all (e.g executable not found), the
    results list will contain the Exception object instead of a
    SubprocessResult.

    You can find this list of results in context `cmdOut` after step completes.

    context['cmds'] will interpolate anything in curly braces for values
    found in context. So if your context looks like this:
        key1: value1
        key2: value2
        cmds:
            - mything --arg1 {key1}

    The cmd passed to the o/s will be "mything --arg value1"
    """
    logger.debug("started")

    AsyncCmdStep(name=__name__, context=context, is_shell=False).run_step()

    logger.debug("done")



"""Nested calls inside loops. These pipelines are in ./tests/pipelines/."""
import tests.common.pipeline_runner as test_pipe_runner
# ------------------------- for ----------------------------------------------#

expected_file_name = '{0}_expected_output.txt'


def test_pipeline_nested_for():
    """Nested jump control-of-flow works with for loop."""
    pipename = 'tests/pipelines/jump/nestedfor'
    expected = expected_file_name.format(pipename)
    test_pipe_runner.assert_pipeline_notify_match_file(pipename, expected)


# ------------------------- for ----------------------------------------------#



"""moduleloader.py unit tests."""
from pathlib import Path
import sys
from unittest.mock import patch

import pytest

from pypyr.errors import PyModuleNotFoundError
import pypyr.moduleloader as moduleloader

# region ImportVisitor


def test_import_visitor_empty():
    """Empty source parses to empty namespace."""
    assert moduleloader.ImportVisitor().get_namespace('') == {}


def test_import_visitor():
    """Import visitor parses python import/import from syntax variations.

    Do NOT import tests.arbpack.arbmod4 anywhere else in the code.
    """
    source = """\
import operator
import itertools as itools

import urllib.parse
import tests.arbpack.arbmod as z

# from mod import submod
from tests.arbpack import arbmod2
from tests.arbpack import arbmod3 as ab3
from tests.arbpack import arbmod4_avoid

# from mod import attr
from decimal import Decimal
from fractions import Fraction as myfraction

from math import ceil, floor
from tests.arbpack.arbmultiattr import arb_attr as x, arb_func as y
"""

    visitor = moduleloader.ImportVisitor()
    ns = visitor.get_namespace(source)

    assert len(ns) == 13

    exec_me = """\
preexisting = 'updated'
arb = len('hello')

modded = operator.mod(6, 4)
prod = list(itools.product('AB', 'ab'))
urlhost = urllib.parse.urlparse('http://arbhost/blah').netloc
z.arbmod_attribute()
ab2 = arbmod2.arb_func_in_arbmod2('ab2 value')
ab3_out = ab3.arb_func_in_arbmod3(123)
ab4 = arbmod4_avoid.arb_func_in_arbmod4(True)
dec = int(Decimal(4).sqrt())
frac = myfraction(1, 3).denominator

ceiling = ceil(6.1)
thefloor = floor(6.1)
arb_attr_out = x
func_res = y('test me')
    """

    locals = {'preexisting': 'initial value'}
    exec(exec_me, ns, locals)

    assert locals['preexisting'] == 'updated'
    assert locals['arb'] == 5
    assert locals['modded'] == 2
    assert locals['prod'] == [('A', 'a'), ('A', 'b'), ('B', 'a'), ('B', 'b')]
    assert locals['urlhost'] == 'arbhost'
    assert locals['ab2'] == 'ab2 value'
    assert locals['ab4'] is True
    assert locals['ab3_out'] == 123
    assert locals['dec'] == 2
    assert locals['frac'] == 3
    assert locals['ceiling'] == 7
    assert locals['thefloor'] == 6
    assert locals['arb_attr_out'] == 123.456
    assert locals['func_res'] == 'test me'


def test_import_visitor_repeating_parent():
    """Repeating parent only shows up 1x in namespace."""
    source = """\
import tests.arbpack.arbmod
import tests.arbpack.arbmod2
import tests.arbpack.arbmod3
"""

    ns = moduleloader.ImportVisitor().get_namespace(source)

    assert len(ns) == 1
    # just see it passes, no return value to assert
    ns['tests'].arbpack.arbmod.arbmod_attribute()
    assert ns['tests'].arbpack.arbmod2.arb_func_in_arbmod2('ab2') == 'ab2'
    assert ns['tests'].arbpack.arbmod3.arb_func_in_arbmod3('ab3') == 'ab3'

    # parent did not import anything NOT specified.
    # tests.arbpack.arbstep exists but wasn't specified for import.
    assert not hasattr(ns['tests'].arbpack, 'arbstep')


def test_import_visitor_relative_raises():
    """Relative imports not supported."""
    source = "from .errors_test import arbthing"

    visitor = moduleloader.ImportVisitor()
    with pytest.raises(TypeError) as err:
        visitor.get_namespace(source)

    assert str(err.value) == ("you can't use relative imports here. use "
                              "absolute imports instead.")

# endregion ImportVisitor

# region get_module


@pytest.fixture()
def known_dirs():
    """Do setup and teardown _known_dirs."""
    moduleloader._known_dirs.clear()
    yield
    moduleloader._known_dirs.clear()


def test_get_module_raises():
    """On get_module ModuleNotFoundError on module not found."""
    with pytest.raises(PyModuleNotFoundError) as err:
        moduleloader.get_module('unlikelyblahmodulenameherexxssz')

    assert str(err.value) == (
        "unlikelyblahmodulenameherexxssz.py should be in your pipeline dir, "
        "or in your working dir, or it should be installed in the current "
        "python env."
        "\nIf you have 'package.sub.mod' your pipeline "
        "dir should contain ./package/sub/mod.py\n"
        "If you specified 'mymodulename', your pipeline "
        "dir should contain ./mymodulename.py\n"
        "If the module is not in your pipeline dir nor in the current working "
        "dir, it must exist in your current python env - so you "
        "should have run pip install or setup.py")


def test_get_module_raises_compatible_error():
    """get_module should raise error compatible with ModuleNotFoundError."""
    with pytest.raises(ModuleNotFoundError):
        moduleloader.get_module('unlikelyblahmodulenameherexxssz')


@patch.object(sys, 'path', ['arb'])
def test_get_module_raises_friendly_on_package_import(known_dirs):
    """get_module should not obscure missing module in existing package."""
    p = Path.cwd().joinpath('tests')
    moduleloader.add_sys_path(p)

    with pytest.raises(PyModuleNotFoundError) as err:
        moduleloader.get_module('arbpack.idontexist')

    assert str(err.value) == (
        "arbpack.idontexist.py should be in your pipeline dir, or in your "
        "working dir, or it should be installed in the current python env."
        "\nIf you have 'package.sub.mod' your pipeline "
        "dir should contain ./package/sub/mod.py\n"
        "If you specified 'mymodulename', your pipeline "
        "dir should contain ./mymodulename.py\n"
        "If the module is not in your pipeline dir nor in the current working "
        "dir, it must exist in your current python env - so you "
        "should have run pip install or setup.py")

    assert sys.path == ['arb', str(p)]


@patch.object(sys, 'path', ['arb'])
def test_get_module_raises_on_inner_import(known_dirs):
    """get_module should not hide failing import statements in imported mod."""
    p = Path.cwd().joinpath('tests')
    moduleloader.add_sys_path(p)

    with pytest.raises(PyModuleNotFoundError) as err:
        moduleloader.get_module('arbpack.arbinvalidimportmod')

    assert str(err.value) == (
        'error importing module blahblah in arbpack.arbinvalidimportmod')

    assert sys.path == ['arb', str(p)]


@patch.object(sys, 'path', ['arb'])
def test_get_module_pass(known_dirs):
    """Pass when get_module finds a module in cwd."""
    p = Path.cwd().joinpath('tests', 'testfiles')
    moduleloader.add_sys_path(p)

    arb_module = moduleloader.get_module('arb')

    assert arb_module
    assert arb_module.__name__ == 'arb'
    assert hasattr(arb_module, 'arb_attribute')

    assert sys.path == ['arb', str(p)]


@patch.object(sys, 'path', ['arb'])
def test_get_module_in_package_pass(known_dirs):
    """See get_module find a module in a package in cwd using dot notation."""
    p = Path.cwd().joinpath('tests')
    moduleloader.add_sys_path(p)
    arb_module = moduleloader.get_module('arbpack.arbmod')

    assert arb_module
    assert arb_module.__name__ == 'arbpack.arbmod'
    assert hasattr(arb_module, 'arbmod_attribute')

    assert sys.path == ['arb', str(p)]

# endregion get_module

# region add_sys_path


@patch.object(sys, 'path', ['arb'])
def test_add_sys_path_nonexisting_dir(known_dirs):
    """Dir not added to sys paths if not exist."""
    p = '/arb/path'
    assert sys.path == ['arb']
    moduleloader.add_sys_path(p)
    assert sys.path == ['arb']
    assert moduleloader._known_dirs == {p}


@patch.object(sys, 'path', ['arb'])
def test_add_sys_path_str(known_dirs):
    """Existing dir added to sys paths."""
    p = 'tests/arbpack'
    assert sys.path == ['arb']
    moduleloader.add_sys_path(p)
    assert sys.path == ['arb', str(Path(p))]
    assert moduleloader._known_dirs == {p}


@patch.object(sys, 'path', ['arb'])
def test_add_sys_path_object(known_dirs):
    """Add sys path takes pathlib Path."""
    p = Path('tests/arbpack')
    assert sys.path == ['arb']
    moduleloader.add_sys_path(p)
    assert sys.path == ['arb', str(p)]
    assert moduleloader._known_dirs == {p}


@patch.object(sys, 'path', ['arb'])
def test_add_sys_path_none(known_dirs):
    """None raises TypeError when trying to add to sys paths."""
    assert None not in sys.path
    with pytest.raises(TypeError):
        moduleloader.add_sys_path(None)

    assert None not in sys.path
    assert sys.path == ['arb']
    assert moduleloader._known_dirs == set()


def assert_list_of_paths_equal(obj, other):
    """Cross platform compare of list of paths."""
    assert [Path(p) for p in obj] == [Path(p) for p in other]


@patch.object(sys, 'path', ['arb'])
def test_add_sys_path_no_dupes(known_dirs):
    """Can't add duplicates to sys path."""
    existing_path = 'tests/arbpack'
    moduleloader.add_sys_path(existing_path)
    assert_list_of_paths_equal(sys.path, ['arb', existing_path])
    moduleloader.add_sys_path(existing_path)
    moduleloader.add_sys_path(existing_path)
    assert sys.path == ['arb', str(Path(existing_path))]
    assert_list_of_paths_equal(sys.path, ['arb', existing_path])
    assert moduleloader._known_dirs == {'tests/arbpack'}


def test_add_sys_path_unknown_but_in_sys_path_already(known_dirs):
    """Unknown dir but it's already in sys.path."""
    moduleloader._known_dirs.clear()
    p = Path.cwd().joinpath('tests')

    with patch.object(sys, 'path', [str(p)]):
        moduleloader.add_sys_path(p)
        assert sys.path == [str(p)]

    assert moduleloader._known_dirs == {p}

# endregion add_sys_path



"""pipelinerunner.py integration tests."""
import logging
from pathlib import Path
from unittest.mock import call

import pytest

from pypyr.cache.loadercache import loader_cache
from pypyr.errors import KeyNotInContextError
from pypyr import pipelinerunner

from tests.common.utils import patch_logger

working_dir_tests = Path.cwd().joinpath('tests')


@pytest.fixture
def pipeline_cache_reset():
    """Invoke for every test function in the module."""
    loader_cache.clear()
    yield
    loader_cache.clear()

# region smoke


def test_pipeline_runner_main(pipeline_cache_reset):
    """Smoke test for pipeline runner main.

    Strictly speaking this is an integration test, not a unit test.
    """
    pipelinerunner.run(pipeline_name='tests/pipelines/smoke',
                       args_in=None,
                       py_dir=working_dir_tests)
# endregion smoke

# region main


def test_pipeline_runner_main_all(pipeline_cache_reset):
    """Run main with all arguments as expected."""
    expected_notify_output = ['sg1', 'sg1.2', 'success_handler']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        pipelinerunner.run(
            # this path works because naiveloader checks from tests subdir.
            pipeline_name='pipelines/api/main-all',
            args_in=['A', 'B', 'C'],
            groups=['sg1'],
            success_group='sh',
            failure_group='fh',
            loader='arbpack.naivefileloader',
            py_dir=working_dir_tests)

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_all_with_failure(pipeline_cache_reset):
    """Run main with all arguments as expected with runtime error."""
    expected_notify_output = ['sg2', 'success_handler', 'fh']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        with pytest.raises(ValueError) as err:
            pipelinerunner.run(
                pipeline_name='pipelines/api/main-all',
                args_in=['A', 'B', 'C', 'raise on sh'],
                groups=['sg2'],
                success_group='sh',
                failure_group='fh',
                loader='arbpack.naivefileloader',
                py_dir=working_dir_tests)

    assert str(err.value) == "err from sh"
    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_minimal():
    """Run main with minimal arguments as expected."""
    expected_notify_output = ['steps', 'argList==None', 'on_success']

    # working_dir will default to repo root rather than test root
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        pipelinerunner.run('tests/pipelines/api/main-all')

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_with_failure():
    """Run main with failure argument as expected."""
    expected_notify_output = ['sg3', 'fh']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        with pytest.raises(ValueError) as err:
            pipelinerunner.run(
                pipeline_name='tests/pipelines/api/main-all',
                groups=['sg3'],
                failure_group='fh')

    assert str(err.value) == "err from sg3"
    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_minimal_with_failure_handled():
    """Run main minimal with failure argument as expected."""
    expected_notify_output = ['steps', 'on_success', 'on_failure']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        pipelinerunner.run(
            pipeline_name='tests/pipelines/api/main-all',
            args_in=['A', 'B', 'C', 'raise on success'])

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_with_failure_handled():
    """Run main with failure argument as expected."""
    expected_notify_output = ['sg3', 'on_failure']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        pipelinerunner.run(pipeline_name='tests/pipelines/api/main-all',
                           groups=['sg3'],
                           failure_group='on_failure')

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]

# endregion main

# region main_with_context


def test_pipeline_runner_main_with_context_all(pipeline_cache_reset):
    """Run main with context with all arguments as expected."""
    expected_notify_output = ['sg1', 'sg1.2', 'success_handler']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        out = pipelinerunner.run(
            pipeline_name='pipelines/api/main-all',
            dict_in={'argList': ['A', 'B', 'C']},
            # parse_args=False,
            groups=['sg1'],
            success_group='sh',
            failure_group='fh',
            loader='arbpack.naivefileloader',
            py_dir=working_dir_tests)

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]
    assert out == {'argList': ['A', 'B', 'C'], 'set_in_pipe': 123}


def test_pipeline_runner_main_with_context_all_with_failure(
        pipeline_cache_reset):
    """Run main with context - all arguments as expected with runtime error."""
    expected_notify_output = ['sg2', 'success_handler', 'fh']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        with pytest.raises(ValueError) as err:
            pipelinerunner.run(
                pipeline_name='pipelines/api/main-all',
                dict_in={'argList': ['A', 'B', 'C', 'raise on sh']},
                groups=['sg2'],
                success_group='sh',
                failure_group='fh',
                loader='arbpack.naivefileloader',
                py_dir=working_dir_tests)

    assert str(err.value) == "err from sh"
    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_with_context_minimal():
    """Run main with context with minimal arguments as expected."""
    # Not having argList==None proves context_parser didn't run.
    expected_notify_output = ['steps', 'argList==None', 'on_success']

    # working_dir will default to repo root rather than test root
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        out = pipelinerunner.run('tests/pipelines/api/main-all')

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]

    assert out == {'argList': [], 'set_in_pipe': 456}
    # somewhat arbitrary check if behaves like Context()
    out.assert_key_has_value('set_in_pipe', 'caller')


def test_pipeline_runner_main_with_context_with_failure():
    """Run main with context with failure argument as expected."""
    expected_notify_output = ['sg3', 'fh']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        with pytest.raises(ValueError) as err:
            pipelinerunner.run(
                pipeline_name='tests/pipelines/api/main-all',
                groups=['sg3'],
                failure_group='fh')

    assert str(err.value) == "err from sg3"
    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]


def test_pipeline_runner_main_with_context_relative_working_dir(
        pipeline_cache_reset):
    """Run main with context with relative working directory."""
    expected_notify_output = ['steps', 'on_success', 'on_failure']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        out = pipelinerunner.run(
            pipeline_name='tests/pipelines/api/main-all',
            dict_in={'argList': ['A', 'B', 'C', 'raise on success']})

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]

    assert len(out) == 4
    assert out['argList'] == ['A', 'B', 'C', 'raise on success']
    assert out['set_in_pipe'] == 456
    assert out['py'] == "raise ValueError('err from on_success')"

    assert len(out['runErrors']) == 1
    out_run_error = out['runErrors'][0]
    assert out_run_error
    assert out_run_error['col'] == 5
    assert out_run_error['customError'] == {}
    assert out_run_error['description'] == 'err from on_success'
    assert repr(out_run_error['exception']) == repr(ValueError(
        'err from on_success'))
    assert out_run_error['line'] == 74
    assert out_run_error['name'] == 'ValueError'
    assert out_run_error['step'] == 'pypyr.steps.py'
    assert out_run_error['swallowed'] is False
    # somewhat arbitrary check if behaves like Context()
    out.assert_key_has_value('set_in_pipe', 'caller')


def test_pipeline_runner_main_with_context_minimal_with_failure_handled():
    """Run main with context minimal with failure argument as expected."""
    expected_notify_output = ['steps', 'on_success', 'on_failure']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        out = pipelinerunner.run(
            pipeline_name='tests/pipelines/api/main-all',
            dict_in={'argList': ['A', 'B', 'C', 'raise on success']})

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]

    assert len(out) == 4
    assert out['argList'] == ['A', 'B', 'C', 'raise on success']
    assert out['set_in_pipe'] == 456
    assert out['py'] == "raise ValueError('err from on_success')"

    assert len(out['runErrors']) == 1
    out_run_error = out['runErrors'][0]
    assert out_run_error
    assert out_run_error['col'] == 5
    assert out_run_error['customError'] == {}
    assert out_run_error['description'] == 'err from on_success'
    assert repr(out_run_error['exception']) == repr(ValueError(
        'err from on_success'))
    assert out_run_error['line'] == 74
    assert out_run_error['name'] == 'ValueError'
    assert out_run_error['step'] == 'pypyr.steps.py'
    assert out_run_error['swallowed'] is False
    # somewhat arbitrary check if behaves like Context()
    out.assert_key_has_value('set_in_pipe', 'caller')


def test_pipeline_runner_main_with_context_with_failure_handled():
    """Run main with context with failure argument as expected."""
    expected_notify_output = ['sg3', 'on_failure']
    with patch_logger('pypyr.steps.echo', logging.NOTIFY) as mock_log:
        out = pipelinerunner.run(
            pipeline_name='tests/pipelines/api/main-all',
            groups=['sg3'],
            failure_group='on_failure')

    assert mock_log.mock_calls == [call(v) for v in expected_notify_output]

    # extra context items is argList: []
    assert len(out) == 3
    assert out['py'] == "raise ValueError('err from sg3')"

    assert len(out['runErrors']) == 1
    out_run_error = out['runErrors'][0]
    assert out_run_error
    assert out_run_error['col'] == 5
    assert out_run_error['customError'] == {}
    assert out_run_error['description'] == 'err from sg3'
    assert repr(out_run_error['exception']) == repr(ValueError('err from sg3'))
    assert out_run_error['line'] == 50
    assert out_run_error['name'] == 'ValueError'
    assert out_run_error['step'] == 'pypyr.steps.py'
    assert out_run_error['swallowed'] is False

    # somewhat arbitrary check if behaves like Context()
    with pytest.raises(KeyNotInContextError) as err:
        out.assert_key_has_value('set_in_pipe', 'arbcaller')

    assert str(err.value) == ("context['set_in_pipe'] doesn't exist. It must "
                              "exist for arbcaller.")

# endregion main_with_context
